import datetime as dt
import re
from typing import Dict, Optional
from urllib.error import HTTPError, URLError

import pandas as pd
import requests
from bs4 import BeautifulSoup
from opnieuw import retry
from pandas.tseries.offsets import MonthEnd
from requests.exceptions import ConnectionError

from econuy.retrieval.external_sector import reserves
from econuy import transform
from econuy.core import Pipeline
from econuy.utils import metadata
from econuy.utils.sources import urls
from econuy.utils.extras import fiscal_sheets, taxes_columns


@retry(
    retry_on_exceptions=(HTTPError, ConnectionError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def _balance_retriever() -> Dict[str, pd.DataFrame]:
    """Helper function. See any of the `balance_...()` functions."""
    response = requests.get(urls["balance_gps"]["dl"]["main"])
    soup = BeautifulSoup(response.content, "html.parser")
    links = soup.find_all(href=re.compile("\\.xlsx$"))
    link = links[0]["href"]
    xls = pd.ExcelFile(link)
    output = {}
    for dataset, meta in fiscal_sheets.items():
        data = (pd.read_excel(xls, sheet_name=meta["sheet"]).
                dropna(axis=0, thresh=4).dropna(axis=1, thresh=4).
                transpose().set_index(2, drop=True))
        data.columns = data.iloc[0]
        data = data[data.index.notnull()].rename_axis(None)
        data.index = data.index + MonthEnd(1)
        data.columns = meta["colnames"]
        data = data.apply(pd.to_numeric, errors="coerce")
        data.rename_axis(None, inplace=True)
        metadata._set(
            data, area="Sector público", currency="UYU",
            inf_adj="No", unit="Millones", seas_adj="NSA",
            ts_type="Flujo", cumperiods=1
        )

        output.update({dataset: data})

    return output


def balance_gps() -> pd.DataFrame:
    """Get fiscal balance data for the consolidated public sector.

    Returns
    -------
    Monthly fiscal balance for the consolidated public sector : pd.DataFrame

    """
    return _balance_retriever()["gps"]


def balance_nfps() -> pd.DataFrame:
    """Get fiscal balance data for the non-financial public sector.

    Returns
    -------
    Monthly fiscal balance for the non-financial public sector : pd.DataFrame

    """
    return _balance_retriever()["nfps"]


def balance_cg_bps() -> pd.DataFrame:
    """Get fiscal balance data for the central government + BPS.

    Returns
    -------
    Monthly fiscal balance for the central government + BPS : pd.DataFrame

    """
    return _balance_retriever()["cg-bps"]


def balance_pe() -> pd.DataFrame:
    """Get fiscal balance data for public enterprises.

    Returns
    -------
    Monthly fiscal balance for public enterprises : pd.DataFrame

    """
    return _balance_retriever()["pe"]


def balance_ancap() -> pd.DataFrame:
    """Get fiscal balance data for ANCAP.

    Returns
    -------
    Monthly fiscal balance for ANCAP : pd.DataFrame

    """
    return _balance_retriever()["ancap"]


def balance_ute() -> pd.DataFrame:
    """Get fiscal balance data for UTE.

    Returns
    -------
    Monthly fiscal balance for UTE : pd.DataFrame

    """
    return _balance_retriever()["ute"]


def balance_antel() -> pd.DataFrame:
    """Get fiscal balance data for ANTEL.

    Returns
    -------
    Monthly fiscal balance for ANTEL : pd.DataFrame

    """
    return _balance_retriever()["antel"]


def balance_ose() -> pd.DataFrame:
    """Get fiscal balance data for OSE.

    Returns
    -------
    Monthly fiscal balance for OSE : pd.DataFrame

    """
    return _balance_retriever()["ose"]


@retry(
    retry_on_exceptions=(HTTPError, ConnectionError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def tax_revenue() -> pd.DataFrame:
    """
    Get tax revenues data.

    This retrieval function requires that Ghostscript and Tkinter be found in
    your system.

    Returns
    -------
    Monthly tax revenues : pd.DataFrame

    """
    raw = pd.read_excel(urls["tax_revenue"]["dl"]["main"],
                        usecols="C:AO", index_col=0)
    raw.index = pd.to_datetime(raw.index, errors="coerce")
    output = raw.loc[~pd.isna(raw.index)]
    output.index = output.index + MonthEnd(0)
    output.columns = taxes_columns
    output = output.div(1000000)
    latest = pd.read_csv(urls["tax_revenue"]["dl"]["pdfs"],
                         index_col=0, parse_dates=True)
    latest = latest.loc[[x not in output.index for x in latest.index]]
    for col in latest.columns:
        for date in latest.index:
            prev_year = date  + MonthEnd(-12)
            output.loc[date, col] = (output.loc[prev_year, col]
                                     * (1 + latest.loc[date, col] / 100))
    output = pd.concat([output, latest], sort=False)
    output = output.loc[~output.index.duplicated(keep="first")]

    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Sector público", currency="UYU",
                  inf_adj="No", unit="Millones", seas_adj="NSA",
                  ts_type="Flujo", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=30,
)
def _public_debt_retriever() -> Dict[str, pd.DataFrame]:
    """Helper function. See any of the `public_debt_...()` functions."""
    colnames = ["Total deuda", "Plazo contractual: hasta 1 año",
                "Plazo contractual: entre 1 y 5 años",
                "Plazo contractual: más de 5 años",
                "Plazo residual: hasta 1 año",
                "Plazo residual: entre 1 y 5 años",
                "Plazo residual: más de 5 años",
                "Moneda: pesos", "Moneda: dólares", "Moneda: euros",
                "Moneda: yenes", "Moneda: DEG", "Moneda: otras",
                "Residencia: no residentes", "Residencia: residentes"]

    xls = pd.ExcelFile(urls["public_debt_gps"]["dl"]["main"])
    gps_raw = pd.read_excel(xls, sheet_name="SPG2",
                            usecols="B:Q", index_col=0, skiprows=10,
                            nrows=(dt.datetime.now().year - 1999) * 4)
    gps = gps_raw.dropna(how="any", thresh=2)
    gps.index = pd.date_range(start="1999-12-31", periods=len(gps),
                              freq="Q-DEC")
    gps.columns = colnames

    nfps_raw = pd.read_excel(xls, sheet_name="SPNM bruta",
                             usecols="B:O", index_col=0)
    loc = nfps_raw.index.get_loc("9. Deuda Bruta del Sector Público no "
                                 "monetario por plazo y  moneda.")
    nfps = nfps_raw.iloc[loc + 5:, :].dropna(how="any")
    nfps.index = pd.date_range(start="1999-12-31", periods=len(nfps),
                               freq="Q-DEC")
    nfps_extra_raw = pd.read_excel(xls, sheet_name="SPNM bruta",
                                   usecols="O:P", skiprows=11,
                                   nrows=(dt.datetime.now().year - 1999) * 4)
    nfps_extra = nfps_extra_raw.dropna(how="all")
    nfps_extra.index = nfps.index
    nfps = pd.concat([nfps, nfps_extra], axis=1)
    nfps.columns = colnames

    cb_raw = pd.read_excel(xls, sheet_name="BCU bruta",
                           usecols="B:O", index_col=0,
                           skiprows=(dt.datetime.now().year - 1999) * 8 + 20)
    cb = cb_raw.dropna(how="any")
    cb.index = pd.date_range(start="1999-12-31", periods=len(cb),
                             freq="Q-DEC")
    cb_extra_raw = pd.read_excel(xls, sheet_name="BCU bruta",
                                 usecols="O:P", skiprows=11,
                                 nrows=(dt.datetime.now().year - 1999) * 4)
    bcu_extra = cb_extra_raw.dropna(how="all")
    bcu_extra.index = cb.index
    cb = pd.concat([cb, bcu_extra], axis=1)
    cb.columns = colnames

    assets_raw = pd.read_excel(xls, sheet_name="Activos Neta",
                               usecols="B,C,D,K", index_col=0, skiprows=13,
                               nrows=(dt.datetime.now().year - 1999) * 4)
    assets = assets_raw.dropna(how="any")
    assets.index = pd.date_range(start="1999-12-31", periods=len(assets),
                                 freq="Q-DEC")
    assets.columns = ["Total activos", "Sector público no monetario",
                      "BCU"]

    output = {"gps": gps, "nfps": nfps, "cb": cb, "assets": assets}

    for meta, data in output.items():
        data.rename_axis(None, inplace=True)
        metadata._set(data, area="Sector público", currency="USD",
                      inf_adj="No", unit="Millones", seas_adj="NSA",
                      ts_type="Stock", cumperiods=1)

        output.update({meta: data})

    return output


def public_debt_gps() -> pd.DataFrame:
    """Get public debt data for the consolidated public sector.

    Returns
    -------
    Quarterly public debt data for the consolidated public sector: pd.DataFrame

    """
    return _public_debt_retriever()["gps"]


def public_debt_nfps() -> pd.DataFrame:
    """Get public debt data for the non-financial public sector.

    Returns
    -------
    Quarterly public debt data for the non-financial public sector: pd.DataFrame

    """
    return _public_debt_retriever()["nfps"]


def public_debt_cb() -> pd.DataFrame:
    """Get public debt data for the central bank

    Returns
    -------
    Quarterly public debt data for the central bank : pd.DataFrame

    """
    return _public_debt_retriever()["cb"]


def public_assets() -> pd.DataFrame:
    """Get public sector assets data.

    Returns
    -------
    Quarterly public sector assets: pd.DataFrame

    """
    return _public_debt_retriever()["assets"]


def net_public_debt(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """
    Get net public debt excluding deposits at the central bank.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Net public debt excl. deposits at the central bank : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()

    pipeline.get("public_debt_gps")
    gross_debt = pipeline.dataset.loc[:, ["Total deuda"]]
    pipeline.get("public_assets")
    assets = pipeline.dataset.loc[:, ["Total activos"]]
    gross_debt.columns = ["Deuda neta del sector"
                          " público global excl. encajes"]
    assets.columns = gross_debt.columns
    pipeline.get("reserves")
    deposits = pipeline.dataset.loc[:,
                                     ["Obligaciones en ME con el sector financiero"]]
    deposits = (transform.resample(deposits, rule="Q-DEC", operation="last")
                .reindex(gross_debt.index).squeeze())
    output = gross_debt.add(assets).add(deposits, axis=0).dropna()
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Sector público",
                  currency="USD", inf_adj="No", unit="Millones",
                  seas_adj="NSA", ts_type="Stock", cumperiods=1)

    return output


def balance_summary(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """
    Get the summary fiscal balance table found in the `Budget Law
    <https://www.gub.uy/contaduria-general-nacion/sites/
    contaduria-general-nacion/files/2020-09/
    Mensaje%20y%20Exposici%C3%B3n%20de%20motivos.pdf>`_. Includes adjustments
    for the `Social Security Fund <https://www.impo.com.uy/bases/decretos/
    71-2018/25>`_.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Summary fiscal balance table : pd.DataFrame

    """
    if pipeline is None or pipeline.download is True:
        data = _balance_retriever()
        gps = data["gps"]
        nfps = data["nfps"]
        gc = data["cg-bps"]
        pe = data["pe"]
    else:
        pipeline.get("balance_gps")
        gps = pipeline.dataset
        pipeline.get("balance_nfps")
        nfps = pipeline.dataset
        pipeline.get("balance_cg-bps")
        gc = pipeline.dataset
        pipeline.get("balance_pe")
        pe = pipeline.dataset

    proc = pd.DataFrame(index=gps.index)

    proc["Ingresos: GC-BPS"] = gc["Ingresos: GC-BPS"]
    proc["Ingresos: GC-BPS ex. FSS"] = (gc["Ingresos: GC-BPS"]
                                        - gc["Ingresos: FSS - Cincuentones"])
    proc["Ingresos: GC"] = gc["Ingresos: GC"]
    proc["Ingresos: DGI"] = gc["Ingresos: DGI"]
    proc["Ingresos: Comercio ext."] = gc["Ingresos: Comercio ext."]
    proc["Ingresos: Otros"] = (gc["Ingresos: GC"]
                               - gc["Ingresos: DGI"]
                               - gc["Ingresos: Comercio ext."])
    proc["Ingresos: BPS"] = gc["Ingresos: BPS neto"]
    proc["Ingresos: FSS - Cincuentones"] = gc["Ingresos: FSS - Cincuentones"]
    proc["Ingresos: BPS ex FSS"] = (gc["Ingresos: BPS neto"]
                                    - gc["Ingresos: FSS - Cincuentones"])
    proc["Egresos: Primarios GC-BPS"] = (gc["Egresos: GC-BPS"]
                                         - gc["Intereses: Total"])
    proc["Egresos: Primarios corrientes GC-BPS"] = (proc["Egresos: Primarios GC-BPS"]
                                                    - gc["Egresos: Inversión"].squeeze())
    proc["Egresos: Remuneraciones"] = gc["Egresos: Remuneraciones"]
    proc["Egresos: No personales"] = gc["Egresos: No personales"]
    proc["Egresos: Pasividades"] = gc["Egresos: Pasividades"]
    proc["Egresos: Transferencias"] = gc["Egresos: Transferencias"]
    proc["Egresos: Inversión"] = gc["Egresos: Inversión"]
    proc["Resultado: Primario GC-BPS"] = (proc["Ingresos: GC-BPS"]
                                          - proc["Egresos: Primarios GC-BPS"])
    proc["Resultado: Primario GC-BPS ex FSS"] = (proc["Ingresos: GC-BPS ex. FSS"]
                                                 - proc["Egresos: Primarios GC-BPS"])
    proc["Intereses: GC-BPS"] = gc["Intereses: Total"]
    proc["Intereses: FSS - Cincuentones"] = gc["Intereses: FSS - Cincuentones"]
    proc["Intereses: GC-BPS ex FSS"] = (proc["Intereses: GC-BPS"]
                                        - proc["Intereses: FSS - Cincuentones"])
    proc["Resultado: Global GC-BPS"] = (proc["Resultado: Primario GC-BPS"]
                                        - proc["Intereses: GC-BPS"])
    proc["Resultado: Global GC-BPS ex FSS"] = (proc["Resultado: Primario GC-BPS ex FSS"]
                                               - proc["Intereses: GC-BPS ex FSS"])

    proc["Resultado: Primario corriente EEPP"] = nfps["Ingresos: Res. primario corriente EEPP"]
    proc["Egresos: Inversiones EEPP"] = pe["Egresos: Inversiones"]
    proc["Resultado: Primario EEPP"] = (proc["Resultado: Primario corriente EEPP"]
                                        - proc["Egresos: Inversiones EEPP"])
    proc["Intereses: EEPP"] = pe["Intereses"]
    proc["Resultado: Global EEPP"] = (proc["Resultado: Primario EEPP"]
                                      - proc["Intereses: EEPP"])

    proc["Resultado: Primario intendencias"] = nfps["Resultado: Primario intendencias"]
    proc["Intereses: Intendencias"] = nfps["Intereses: Intendencias"]
    proc["Resultado: Global intendencias"] = (proc["Resultado: Primario intendencias"]
                                              - proc["Intereses: Intendencias"])

    proc["Resultado: Primario BSE"] = nfps["Resultado: Primario BSE"]
    proc["Intereses: BSE"] = nfps["Intereses: BSE"]
    proc["Resultado: Global BSE"] = (proc["Resultado: Primario BSE"]
                                     - proc["Intereses: BSE"])

    proc["Resultado: Primario resto SPNF"] = (proc["Resultado: Primario EEPP"]
                                              + proc["Resultado: Primario intendencias"]
                                              + proc["Resultado: Primario BSE"])
    proc["Intereses: Resto SPNF"] = (proc["Intereses: EEPP"]
                                     + proc["Intereses: Intendencias"]
                                     + proc["Intereses: BSE"])
    proc["Resultado: Global resto SPNF"] = (proc["Resultado: Global EEPP"]
                                            + proc["Resultado: Global intendencias"]
                                            + proc["Resultado: Global BSE"])
    proc["Resultado: Primario SPNF"] = nfps["Resultado: Primario SPNF"]
    proc["Resultado: Primario SPNF ex FSS"] = (proc["Resultado: Primario SPNF"]
                                               - proc["Ingresos: FSS - Cincuentones"])
    proc["Intereses: SPNF"] = nfps["Intereses: Totales"]
    proc["Intereses: SPNF ex FSS"] = (proc["Intereses: SPNF"]
                                      - proc["Intereses: FSS - Cincuentones"])
    proc["Resultado: Global SPNF"] = nfps["Resultado: Global SPNF"]
    proc["Resultado: Global SPNF ex FSS"] = (proc["Resultado: Primario SPNF ex FSS"]
                                             - proc["Intereses: SPNF ex FSS"])

    proc["Resultado: Primario BCU"] = gps["Resultado: Primario BCU"]
    proc["Intereses: BCU"] = gps["Intereses: BCU"]
    proc["Resultado: Global BCU"] = gps["Resultado: Global BCU"]

    proc["Resultado: Primario SPC"] = gps["Resultado: Primario SPC"]
    proc["Resultado: Primario SPC ex FSS"] = (proc["Resultado: Primario SPNF ex FSS"]
                                              + proc["Resultado: Primario BCU"])
    proc["Intereses: SPC"] = proc["Intereses: SPNF"] + proc["Intereses: BCU"]
    proc["Intereses: SPC ex FSS"] = (proc["Intereses: SPNF ex FSS"]
                                     + proc["Intereses: BCU"])
    proc["Resultado: Global SPC"] = (proc["Resultado: Global SPNF"]
                                     + proc["Resultado: Global BCU"])
    proc["Resultado: Global SPC ex FSS"] = (proc["Resultado: Global SPNF ex FSS"]
                                            + proc["Resultado: Global BCU"])
    output = proc
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Sector público",
                  currency="UYU", inf_adj="No", unit="Millones",
                  seas_adj="NSA", ts_type="Flujo", cumperiods=1)

    return output



from __future__ import annotations
import logging
import copy
import traceback
from datetime import datetime
from inspect import getmodule
from os import PathLike
from pathlib import Path
from typing import Optional, Union, Sequence, Dict, List

import pandas as pd
from sqlalchemy.engine.base import Connection, Engine

from econuy import transform
from econuy.core import Pipeline
from econuy.utils import datasets, logutil, ops
from econuy.utils.exceptions import RetryLimitError


class Session(object):
    """
    A download and transformation session that creates a Pipeline object and
    simplifies working with multiple datasets.

    Alternatively, can be created directly from a Pipeline by using the
    :mod:`~econuy.session.Session.from_pipeline` class method.

    Attributes
    ----------
    location : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
               default None
        Either Path or path-like string pointing to a directory where to find
        a CSV for updating and saving, SQLAlchemy connection or engine object,
        or ``None``, don't save or update.
    download : bool, default True
        If False the ``get`` method will only try to retrieve data on disk.
    always_Save : bool, default True
        If True, save every retrieved dataset to the specified ``location``.
    read_fmt : {'csv', 'xls', 'xlsx'}
        File format of previously downloaded data. Ignored if ``location``
        points to a SQL object.
    save_fmt : {'csv', 'xls', 'xlsx'}
        File format for saving. Ignored if ``location`` points to a SQL object.
    read_header : {'included', 'separate', None}
        Location of dataset metadata headers. 'included' means they are in the
        first 9 rows of the dataset. 'separate' means they are in a separate
        Excel sheet (if ``read_fmt='csv'``, headers are discarded).
        None means there are no metadata headers.
    save_header : {'included', 'separate', None}
        Location of dataset metadata headers. 'included' means they will be set
        as the first 9 rows of the dataset. 'separate' means they will be saved
        in a separate Excel sheet (if ``save_fmt='csv'``, headers are
        discarded). None discards any headers.
    errors : {'raise', 'coerce', 'ignore'}
        How to handle errors that arise from transformations. ``raise`` will
        raise a ValueError, ``coerce`` will force the data into ``np.nan`` and
        ``ignore`` will leave the input data as is.
    log : {str, 0, 1, 2}
        Controls how logging works. ``0``: don't log; ``1``: log to console;
        ``2``: log to console and file with default file; ``str``: log to
        console and file with filename=str
    logger : logging.Logger, default None
        Logger object. For most cases this attribute should be ``None``,
        allowing :attr:`log` to control how logging works.
    max_retries : int, default 3
        Number of retries for :mod:`get` in case any of the selected
        datasets cannot be retrieved.

    """

    def __init__(self,
                 location: Union[str, PathLike,
                                 Connection, Engine, None] = None,
                 download: bool = True,
                 always_save: bool = True,
                 read_fmt: str = "csv",
                 read_header: Optional[str] = "included",
                 save_fmt: str = "csv",
                 save_header: Optional[str] = "included",
                 errors: str = "raise",
                 log: Union[int, str] = 1,
                 logger: Optional[logging.Logger] = None,
                 max_retries: int = 3):
        self.location = location
        self.download = download
        self.always_save = always_save
        self.read_fmt = read_fmt
        self.read_header = read_header
        self.save_fmt = save_fmt
        self.save_header = save_header
        self.errors = errors
        self.log = log
        self.logger = logger
        self.max_retries = max_retries

        self._datasets = {}
        self._retries = 1

        if logger is not None:
            self.log = "custom"
        else:
            if isinstance(log, int) and (log < 0 or log > 2):
                raise ValueError("'log' takes either 0 (don't log info),"
                                 " 1 (log to console), 2 (log to console and"
                                 " default file), or str (log to console and "
                                 "file with filename=str).")
            elif log == 2:
                logfile = Path(self.location) / "info.log"
                log_obj = logutil.setup(file=logfile)
            elif isinstance(log, str) and log != "custom":
                logfile = (Path(self.location) / log).with_suffix(".log")
                log_obj = logutil.setup(file=logfile)
            elif log == 1:
                log_obj = logutil.setup(file=None)
            else:
                log_obj = logutil.setup(null=True)
            self.logger = log_obj

    @classmethod
    def from_pipeline(cls, pipeline: Pipeline) -> Session:
        # Alternative constructor
        s = Session(location=pipeline.location, download=pipeline.download,
                    always_save=pipeline.location, read_fmt=pipeline.read_fmt,
                    read_header=pipeline.read_header, save_fmt=pipeline.read_fmt,
                    save_header=pipeline.save_header, errors=pipeline.errors)
        if not pipeline.dataset.empty:
            s._datasets = {pipeline.name: pipeline.dataset}
        return s

    @property
    def pipeline(self) -> Pipeline:
        # Define a property so that changes to Session attributes are passed
        # down to the Pipeline and taken into account in get().
        p = Pipeline(location=self.location, download=self.download,
                     always_save=self.always_save, read_fmt=self.read_fmt,
                     read_header=self.read_header, save_fmt=self.save_fmt,
                     save_header=self.save_header, errors=self.errors)
        return p

    @property
    def datasets(self) -> Dict[str, pd.DataFrame]:
        """Holds retrieved datasets.

        Returns
        -------
        Datasets : Dict[str, pd.DataFrame]
        """
        return self._datasets

    @property
    def datasets_flat(self) -> Dict[str, pd.DataFrame]:
        """Holds retrieved datasets.

        Returns
        -------
        Datasets : Dict[str, pd.DataFrame]
        """
        nometa = copy.deepcopy(self._datasets)
        for v in nometa.values():
            v.columns = v.columns.get_level_values(0)
        return nometa

    def __repr__(self):
        return (f"Session(location={self.location})\n"
                f"Current dataset(s): {list(self.datasets.keys())}")

    def copy(self, deep: bool = False) -> Session:
        """Copy or deepcopy a Session object.

        Parameters
        ----------
        deep : bool, default True
            If True, deepcopy.

        Returns
        -------
        :class:`~econuy.session.Session`

        """
        if deep:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    @staticmethod
    def available_datasets(functions: bool = False) -> Dict[str, Dict]:
        """Return available ``dataset`` arguments for use in
        :mod:`~econuy.session.Session.get`.

        Returns
        -------
        Dataset : Dict[str, Dict]
        """
        original = datasets.original()
        custom = datasets.custom()

        original_final, custom_final = {}, {}
        for d, f in zip([original, custom], [original_final, custom_final]):
            for k, v in d.items():
                if not functions:
                    f.update({k: v["description"]})
                else:
                    f.update({k: v})

        output = {"original": original_final, "custom": custom_final}
        aux = copy.deepcopy(output)
        for k in aux["custom"].keys():
            # Avoid auxiliary datasets in Session methods (like _lin_gdp)
            if k.startswith("_"):
                output["custom"].pop(k, None)
        return output

    def _select_datasets(self, select: Union[str, int, Sequence[str],
                                             Sequence[int]]) -> List[str]:
        """Generate list of dataset names based on selection.

        Parameters
        ----------
        select : str, int, Sequence[str] or Sequence[int], default "all"
            Datasets in :attr:`datasets` to apply transformation on. Can be
            defined with their names or index position.

        Returns
        -------
        List[str]
            List of dataset names.

        Raises
        ------
        ValueError
            If names and indexes are combined.
        """
        keys = list(self.datasets.keys())
        if isinstance(select, Sequence) and not isinstance(select, str):
            if not all(type(select[i]) == type(select[0])
                       for i in range(len(select))):
                raise ValueError("`select` must be all `int` or all `str`")
            if isinstance(select[0], int):
                proc_select = [keys[i] for i in select]
            else:
                proc_select = [i for i in keys if i in select]
        elif isinstance(select, int):
            proc_select = [keys[select]]
        elif select == "all":
            proc_select = keys
        else:
            proc_select = [select]
        return proc_select

    def _apply_transformation(self,
                              transformation: str,
                              select: Union[str, int, Sequence[str],
                                            Sequence[int]] = "all",
                              **kwargs) -> Dict[str, pd.DataFrame]:
        """Helper method to apply transformations on :attr:`datasets`.

        Parameters
        ----------
        transformation : {'resample', 'chg_diff', 'convert', 'decompose', \
                          'rolling', 'rebase'}
            String representing transformation methods in
            :class:`~econuy.retrieval.core.Pipeline`.
        select : str, int, Sequence[str] or Sequence[int], default "all"
            Datasets in :attr:`datasets` to apply transformations on.

        Returns
        -------
        Transformed datasets : Dict[str, pd.DataFrame]

        """
        p = self.pipeline.copy()
        methods = {"resample": p.resample,
                   "chg_diff": p.chg_diff,
                   "convert": p.convert,
                   "decompose": p.decompose,
                   "rolling": p.rolling,
                   "rebase": p.rebase}

        proc_select = self._select_datasets(select=select)

        new_kwargs = {}
        for k, v in kwargs.items():
            if not isinstance(v, Sequence) or isinstance(v, str):
                new_kwargs.update({k: [v] * len(proc_select)})
            elif len(v) != len(proc_select):
                raise ValueError(f"Wrong number of arguments for '{k}'")
            else:
                new_kwargs.update({k: v})

        output = self.datasets.copy()
        for i, name in enumerate(proc_select):
            current_kwargs = {k: v[i] for k, v in new_kwargs.items()}
            p._dataset = self.datasets[name]
            methods[transformation](**current_kwargs)
            transformed = p.dataset
            output.update({name: transformed})

        return output

    def get(self, names: Union[str, Sequence[str]]):
        """
        Main download method.

        Parameters
        ----------
        names : Union[str, Sequence[str]]
            Dataset to download, see available options in
            :mod:`~econuy.session.Session.available_datasets`. Either a string representing
            a dataset name or a sequence of strings in order to download
            several datasets.

        Raises
        ------
        ValueError
            If an invalid string is found in the ``names`` argument.

        """
        if isinstance(names, str):
            names = [names]
        if any(x not in self.pipeline.available_datasets().keys()
               for x in names):
            raise ValueError("Invalid dataset selected.")

        # Deepcopy the Pipeline so that its dataset attribute is not
        # overwritten each time it's accessed within this method.
        p = self.pipeline.copy()

        failed = []
        not_failed = []
        for name in names:
            try:
                p.get(name=name)
                self._datasets.update({name: p.dataset})
                not_failed.append(name)
            except BaseException:
                traceback.print_exc()
                failed.append(name)
                continue
        if len(failed) > 0:
            if self._retries < self.max_retries:
                self._retries += 1
                self.logger.info(f"Failed to retrieve {', '.join(failed)}. "
                                 f"Retrying (run {self._retries}).")
                self.get(names=failed)
            else:
                self.logger.info(f"Could not retrieve {', '.join(failed)}")
                self._retries = 1
                raise RetryLimitError(f"Maximum retries ({self.max_retries})"
                                      f" reached.")
            self._retries = 1
            return
        self._retries = 1
        return

    def get_bulk(self, names: str):
        """
        Get datasets in bulk.

        Parameters
        ----------
        names : {'all', 'original', 'custom', 'economic_activity', \
                 'prices', 'fiscal_accounts', 'labor', 'external_sector', \
                 'financial_sector', 'income', 'international', 'regional'}
            Type of data to download. `all` gets all available datasets,
            `original` gets all original datatsets and `custom` gets all
            custom datasets. The remaining options get all datasets for that
            area.

        Raises
        ------
        ValueError
            If an invalid string is given to the ``names`` argument.

        """
        valid_datasets = ["all", "original", "custom", "economic_activity",
                          "prices", "fiscal_accounts", "labor", "external_sector",
                          "financial_sector", "income", "international",
                          "regional"]
        if names not in valid_datasets:
            raise ValueError(f"'names' can only be one of "
                             f"{', '.join(valid_datasets)}.")
        available_datasets = self.available_datasets(functions=True)
        original_datasets = list(available_datasets["original"].keys())
        custom_datasets = list(available_datasets["custom"].keys())

        if names == "original":
            self.get(names=original_datasets)
        elif names == "custom":
            self.get(names=custom_datasets)
        elif names == "all":
            self.get(names=original_datasets+custom_datasets)
        else:
            original_area_datasets = []
            for k, v in available_datasets["original"].items():
                if names in getmodule(v["function"]).__name__:
                    original_area_datasets.append(k)
            custom_area_datasets = []
            for k, v in available_datasets["custom"].items():
                if names in getmodule(v["function"]).__name__:
                    custom_area_datasets.append(k)
            self.get(names=original_area_datasets+custom_area_datasets)

        return

    def resample(self, rule: Union[pd.DateOffset, pd.Timedelta, str, List],
                 operation: Union[str, List] = "sum",
                 interpolation: Union[str, List] = "linear",
                 warn: Union[bool, List] = False,
                 select: Union[str, int, Sequence[str],
                               Sequence[int]] = "all"):
        """
        Resample to target frequencies.

        See Also
        --------
        :mod:`~econuy.core.Pipeline.resample`

        """
        output = self._apply_transformation(select=select,
                                            transformation="resample",
                                            rule=rule,
                                            operation=operation,
                                            interpolation=interpolation,
                                            warn=warn)

        self._datasets = output
        return

    def chg_diff(self, operation: Union[str, List] = "chg",
                 period: Union[str, List] = "last",
                 select: Union[str, int, Sequence[str],
                               Sequence[int]] = "all"):
        """
        Calculate pct change or difference.

        See Also
        --------
        :mod:`~econuy.core.Pipeline.chg_diff`.

        """
        output = self._apply_transformation(select=select,
                                            transformation="chg_diff",
                                            operation=operation, period=period)

        self._datasets = output
        return

    def decompose(self, component: Union[str, List] = "both",
                  method: Union[str, List] = "x13",
                  force_x13: Union[bool, List] = False,
                  fallback: Union[str, List] = "loess",
                  trading: Union[bool, List] = True,
                  outlier: Union[bool, List] = True,
                  x13_binary: Union[str, PathLike, List] = "search",
                  search_parents: Union[int, List] = 1,
                  ignore_warnings: Union[bool, List] = True,
                  select: Union[str, int, Sequence[str],
                                Sequence[int]] = "all",
                  **kwargs):
        """
        Apply seasonal decomposition.

        See Also
        --------
        :mod:`~econuy.core.Pipeline.decompose`.

        """
        valid_component = ["seas", "trend"]
        if component not in valid_component:
            raise ValueError(f"Only {', '.join(valid_component)} are allowed."
                             f"See underlying 'decompose'.")

        output = self._apply_transformation(select=select,
                                            transformation="decompose",
                                            component=component,
                                            method=method,
                                            force_x13=force_x13,
                                            fallback=fallback,
                                            trading=trading,
                                            outlier=outlier,
                                            x13_binary=x13_binary,
                                            search_parents=search_parents,
                                            ignore_warnings=ignore_warnings,
                                            **kwargs)

        self._datasets = output
        return

    def convert(self, flavor: Union[str, List],
                start_date: Union[str, datetime, None, List] = None,
                end_date: Union[str, datetime, None, List] = None,
                select: Union[str, int, Sequence[str],
                              Sequence[int]] = "all"):
        """Convert to other units.

        See Also
        --------
        :mod:`~econuy.core.Pipeline.convert`.

        """
        if flavor not in ["usd", "real", "gdp", "pcgdp"]:
            raise ValueError("'flavor' can be one of 'usd', 'real', "
                             "or 'gdp'.")

        if flavor == "real":
            output = self._apply_transformation(select=select,
                                                transformation="convert",
                                                flavor=flavor,
                                                start_date=start_date,
                                                end_date=end_date)
        else:
            output = self._apply_transformation(select=select,
                                                transformation="convert",
                                                flavor=flavor)

        self._datasets = output
        return

    def rebase(self, start_date: Union[str, datetime, List],
               end_date: Union[str, datetime, None, List] = None,
               base: Union[float, List] = 100.0,
               select: Union[str, int, Sequence[str],
                             Sequence[int]] = "all"):
        """
        Scale to a period or range of periods.

        See Also
        --------
        :mod:`~econuy.core.Pipeline.rebase`.

        """
        output = self._apply_transformation(select=select,
                                            transformation="rebase",
                                            start_date=start_date,
                                            end_date=end_date, base=base)
        self._datasets = output
        return

    def rolling(self, window: Union[int, List, None] = None,
                operation: Union[str, List] = "sum",
                select: Union[str, int, Sequence[str],
                              Sequence[int]] = "all"):
        """
        Calculate rolling averages or sums.

        See Also
        --------
        :mod:`~econuy.core.Pipeline.rolling`.

        """
        output = self._apply_transformation(select=select,
                                            transformation="rolling",
                                            window=window,
                                            operation=operation)
        self._datasets = output
        return

    def concat(self, select: Union[str, int, Sequence[str],
                                   Sequence[int]] = "all",
               concat_name: Optional[str] = None,
               force_suffix: bool = False):
        """
        Concatenate datasets in :attr:`datasets` and add as a new dataset.

        Resample to lowest frequency of selected datasets.

        Parameters
        ----------
        select : str, int, Sequence[str] or Sequence[int], default "all"
            Datasets to concatenate.
        concat_name : Optional[str], default None
            Name used as a key for the output dataset. The default None sets
            the name to "concat_{dataset_1_name}_..._{dataset_n_name}".
        force_suffix : bool, default False
            Whether to include each dataset's full name as a prefix in all
            indicator columns.

        """
        proc_select = self._select_datasets(select=select)
        proc_select = [x for x in proc_select if "concat_" not in x]
        selected_datasets = [d for n, d in self.datasets.items()
                             if n in proc_select]

        indicator_names = [col for df in selected_datasets
                           for col in df.columns.get_level_values(0)]
        if (len(indicator_names) > len(set(indicator_names))
                or force_suffix is True):
            for n, d in zip(proc_select, selected_datasets):
                try:
                    full_name = self.available_datasets()["original"][n]
                except KeyError:
                    full_name = self.available_datasets()["custom"][n]
                d.columns = d.columns.set_levels(
                    f"{full_name}_" + d.columns.get_level_values(0),
                    level=0
                )

        freqs = [pd.infer_freq(df.index) for df in selected_datasets]
        if all(freq == freqs[0] for freq in freqs):
            combined = pd.concat(selected_datasets, axis=1)
        else:
            for freq_opt in ["A-DEC", "A", "Q-DEC", "Q",
                             "M", "2W-SUN", "W-SUN"]:
                if freq_opt in freqs:
                    output = []
                    for df in selected_datasets:
                        freq_df = pd.infer_freq(df.index)
                        if freq_df == freq_opt:
                            df_match = df.copy()
                        else:
                            type_df = df.columns.get_level_values("Tipo")[0]
                            unit_df = df.columns.get_level_values("Unidad")[0]
                            if type_df == "Stock":
                                df_match = transform.resample(df, rule=freq_opt,
                                                              operation="last")
                            elif (type_df == "Flujo" and
                                  not any(x in unit_df for
                                          x in ["%", "=", "Cambio"])):
                                df_match = transform.resample(df, rule=freq_opt,
                                                              operation="sum")
                            else:
                                df_match = transform.resample(df, rule=freq_opt,
                                                              operation="mean")
                        output.append(df_match)
                    combined = pd.concat(output, axis=1)
                    break
                else:
                    continue
        if concat_name is None:
            concat_name = "concat_" + "_".join(proc_select)
        elif not concat_name.startswith("concat_"):
            concat_name = "concat_" + concat_name
        else:
            pass
        self._datasets.update({concat_name: combined})
        return

    def save(self,
             select: Union[str, int, Sequence[str], Sequence[int]] = "all"):
        """Write datasets.

        Parameters
        ----------
        select : str, int, Sequence[str] or Sequence[int], default "all"
            Datasets to save.

        Raises
        ------
        ValueError
            If `self.location` is None.
        """
        if self.location is None:
            raise ValueError("No save location defined.")

        proc_select = self._select_datasets(select=select)

        for name, dataset in self.datasets.items():
            if name in proc_select:
                ops._io(operation="save", data_loc=self.location,
                        data=dataset, name=name,
                        file_fmt=self.save_fmt,
                        multiindex=self.save_header)
            else:
                continue

        return



from os import path, PathLike, mkdir
from pathlib import Path
from typing import Union, Optional

import pandas as pd
from sqlalchemy.engine.base import Connection, Engine
from sqlalchemy.exc import ProgrammingError, OperationalError

from econuy.utils import metadata, sqlutil


def _read(data_loc: Union[str, PathLike,
                          Connection, Engine],
          file_fmt: str = "csv",
          multiindex: Optional[str] = "included",
          table_name: Optional[str] = None) -> pd.DataFrame:
    """Read a DataFrame from SQL database or CSV/Excel.

    Parameters
    ----------
    data_loc : Union[str, PathLike, Connection, Engine]
        SQL object or path to file.
    table_name : Optional[str]
        Name for input SQL table.
    file_fmt : {'csv', 'xlsx'}
        File format. Ignored if ``data_loc`` refers to a SQL object.
    multiindex : {'included', 'separate', None}
        How to handle multiindexes for metadata. ``None`` keeps only the first
        level, ``included`` keeps as DataFrame columns and ``separate`` saves
        it to another sheet (only valid for Excel-type formats).
    """
    try:
        if isinstance(data_loc, (Engine, Connection)):
            if multiindex is not None:
                previous_data = sqlutil.read(con=data_loc,
                                             table_name=table_name)
            else:
                previous_data = pd.read_sql(sql=table_name,
                                            con=data_loc,
                                            index_col="index",
                                            parse_dates="index")
        else:
            date_format = None
            special_date_format = ["cattle", "reserves_changes"]
            if Path(data_loc).stem in special_date_format:
                date_format = "%d/%m/%Y"
            if file_fmt == "csv":
                if multiindex == "included":
                    previous_data = pd.read_csv(data_loc, index_col=0,
                                                header=list(range(9)),
                                                float_precision="high",
                                                encoding="latin1")
                    metadata._set(previous_data)
                else:
                    previous_data = pd.read_csv(data_loc, index_col=0,
                                                header=0, float_precision="high",
                                                encoding="latin1")
            else:
                if multiindex == "included":
                    previous_data = pd.read_excel(data_loc, index_col=0,
                                                  header=list(range(9)),
                                                  sheet_name="Data")
                    metadata._set(previous_data)
                elif multiindex == "separate":
                    excel = pd.ExcelFile(data_loc)
                    previous_data = pd.read_excel(excel, index_col=0,
                                                  header=0, sheet_name="Data")
                    header = pd.read_excel(excel,
                                           sheet_name="Metadata",
                                           index_col=0).rename_axis("Indicador").T
                    previous_data.columns = pd.MultiIndex.from_frame(header.reset_index())
                    metadata._set(previous_data)
                else:
                    previous_data = pd.read_excel(data_loc, index_col=0,
                                                  header=0, sheet_name="Data")
            try:
                previous_data.index = pd.to_datetime(previous_data.index,
                                                    format=date_format)
            except ValueError:
                previous_data.index = pd.to_datetime(previous_data.index,
                                                     format=None)

    except (ProgrammingError, OperationalError, FileNotFoundError):
        previous_data = pd.DataFrame()

    return previous_data


def _save(data: pd.DataFrame,
          data_loc: Union[str, PathLike, Connection, Engine],
          table_name: str = None, file_fmt: str = "csv",
          multiindex: str = "included"):
    """Save a DataFrame to SQL database or CSV/Excel.

    Parameters
    ----------
    data : pd.DataFrame
        Data to save.
    data_loc : Union[str, PathLike, Connection, Engine]
        SQL object or path to file.
    table_name : Optional[str]
        Name for output SQL table.
    file_fmt : {'csv', 'xlsx'}
        File format. Ignored if ``data_loc`` refers to a SQL object.
    multiindex : {'included', 'separate', None}
        How to handle multiindexes for metadata. ``None`` keeps only the first
        level, ``included`` keeps as DataFrame columns and ``separate`` saves
        it to another sheet (only valid for Excel-type formats).
    """
    if isinstance(data_loc, (Engine, Connection)):
        sqlutil.df_to_sql(data, name=table_name,
                          con=data_loc)
    else:
        data_proc = data.copy()
        if file_fmt == "csv":
            if multiindex != "included":
                data_proc.columns = data_proc.columns.get_level_values(0)
                data_proc.to_csv(data_loc, encoding="latin1")
            else:
                data_proc.to_csv(data_loc, encoding="latin1")
        else:
            if multiindex == "included":
                data_proc.to_excel(data_loc, sheet_name="Data")
            elif multiindex == "separate":
                metadata = data_proc.columns.to_frame().set_index("Indicador").T
                data_proc.columns = data_proc.columns.get_level_values(0)
                with pd.ExcelWriter(data_loc) as f:
                    data_proc.to_excel(f, sheet_name="Data")
                    metadata.to_excel(f, sheet_name="Metadata")
            else:
                data_proc.columns = data_proc.columns.get_level_values(0)
                data_proc.to_excel(data_loc, sheet_name="Data")

    return


def _revise(new_data: pd.DataFrame, prev_data: pd.DataFrame,
            revise_rows: Union[int, str]):
    """Replace n rows of data at the end of a dataframe with new data."""
    if len(prev_data) == 0:
        return new_data
    frequency = pd.infer_freq(prev_data.index)
    freq_table = {"A": 3, "A-DEC": 3, "Q": 4, "Q-DEC": 4, "M": 12}
    new_data = new_data.apply(pd.to_numeric, errors="coerce")

    if isinstance(revise_rows, str) and revise_rows in "noduplicate":
        prev_data.columns = new_data.columns
        updated = prev_data.append(new_data)
        updated = updated.loc[~updated.index.duplicated(keep="last")]
        updated.sort_index(inplace=True)
        return updated

    elif isinstance(revise_rows, str) and revise_rows in "automatic":
        try:
            revise_rows = freq_table[frequency]
        except KeyError:
            revise_rows = 12
            if len(prev_data) <= 12 or len(new_data) <= 12:
                revise_rows = 3

    elif isinstance(revise_rows, int):
        revise_rows = revise_rows
    else:
        raise ValueError("`revise_rows` accepts int, 'nodup' or 'auto'")

    non_revised = prev_data[:len(prev_data) - revise_rows]
    revised = new_data[len(prev_data) - revise_rows:]
    non_revised.columns = new_data.columns
    updated = non_revised.append(revised, sort=False)

    return updated


def _io(operation: str, data_loc: Union[str, PathLike, Connection, Engine],
        name: str, data: Optional[pd.DataFrame] = None,
        file_fmt: str = "csv",
        multiindex: Optional[str] = "included") -> Optional[pd.DataFrame]:
    """Save/read a DataFrame to/from SQL database or CSV/Excel.

    Parameters
    ----------
    operation : {'save', 'read'}
        Whether to save or read.
    data : Optional[pd.DataFrame]
        Data to save.
    data_loc : Union[str, PathLike, Connection, Engine]
        SQL object or path to folder.
    name : str
        Name for SQL table or IO file.
    file_fmt : {'csv', 'xlsx'}
        File format. Ignored if ``data_loc`` refers to a SQL object.
    multiindex : {'included', 'separate', None}
        How to handle multiindexes for metadata. ``None`` keeps only the first
        level, ``included`` keeps as DataFrame columns and ``separate`` saves
        it to another sheet (only valid for Excel-type formats).
    """
    valid_fmt = ["csv", "xlsx", "xls"]
    valid_operation = ["read", "save"]
    if file_fmt not in valid_fmt:
        raise ValueError(f"'file_fmt' must be one of {', '.join(valid_fmt)}.")
    if operation not in valid_operation:
        raise ValueError(f"'operation' must be one of {', '.join(valid_operation)}.")
    if multiindex not in ["included", "separate", None]:
        raise ValueError(f"'multiindex' must be one of 'included', 'separate' or None.")

    suffix = f".{file_fmt}"
    if operation == "read":
        if isinstance(data_loc, (str, PathLike)):
            full_update_loc = (Path(data_loc) / name).with_suffix(suffix)
        else:
            full_update_loc = data_loc
        return _read(full_update_loc, table_name=name,
                     multiindex=multiindex,
                     file_fmt=file_fmt)

    elif operation == "save":
        if isinstance(data_loc, (str, PathLike)):
            full_save_loc = (Path(data_loc) / name).with_suffix(suffix)
            if not path.exists(path.dirname(full_save_loc)):
                mkdir(path.dirname(full_save_loc))
        else:
            full_save_loc = data_loc

        _save(data=data, data_loc=full_save_loc,
              file_fmt=file_fmt, multiindex=multiindex,
              table_name=name)
        return



import logging
from os import listdir, remove, rmdir, path
from pathlib import Path

import pandas as pd
import pytest
from sqlalchemy import create_engine, inspect

from econuy.core import Pipeline
from econuy.session import Session
from econuy.utils import sqlutil, datasets, ops


CUR_DIR = path.abspath(path.dirname(__file__))
TEST_DIR = path.join(CUR_DIR, "test-data")
TEST_CON = create_engine("sqlite://")
sqlutil.insert_csvs(con=TEST_CON, directory=TEST_DIR)


def remove_temporary_files_folders():
    avoid = list(datasets.original().keys()) + list(datasets.custom().keys()) + ["commodity_weights"]
    avoid_csv = [f"{x}.csv" for x in avoid]

    for f in listdir(TEST_DIR):
        if f not in avoid_csv:
            try:
                remove(path.join(TEST_DIR, f))
            except IsADirectoryError:
                rmdir(path.join(TEST_DIR, f))

    for table in inspect(TEST_CON).get_table_names():
        if table not in avoid:
            TEST_CON.engine.execute(f'DROP TABLE IF EXISTS "{table}"')

    return


def test_wrong_dataset():
    p = Pipeline(location=TEST_DIR)
    session = Session.from_pipeline(p)
    for method in [session.get, session.get_bulk]:
        with pytest.raises(ValueError):
            method(names="wrong")


def test_save_no_location():
    s = Session()
    with pytest.raises(ValueError):
        s.save()


@pytest.mark.parametrize("names,fmt,files",
                         [(["cpi"], "xlsx", ["cpi.xlsx"]),
                          (["cpi", "nxr_monthly"], "csv", ["cpi.csv", "nxr_monthly.csv"])])
def test_save(names, fmt, files, tmpdir):
    s = Session(tmpdir, always_save=False, save_fmt=fmt, read_fmt=fmt)
    s.get(names)
    s.save()
    for f in files:
        file_path = Path(tmpdir, f)
        assert path.isfile(file_path)
        assert not ops._read(file_path, file_fmt=fmt).empty


def test_save_different_formats(tmpdir):
    s = Session(tmpdir, read_fmt="xlsx", read_header="separate",
                save_fmt="xlsx", save_header="separate", download=False)
    s.get("cpi") # read: xlsx, separate; save: xlsx, separate
    first = s.datasets["cpi"].copy()

    s.save_fmt = "csv"
    s.save_header = "included"
    s.get("cpi") # read: xlsx, separate; save: csv, included
    second = s.datasets["cpi"].copy()

    s.read_fmt = "csv"
    s.read_header = "included"
    s.save_header = None
    s.get("cpi") # read: csv, included; save: csv, None
    third = s.datasets["cpi"].copy()

    s.read_header = None
    s.get("cpi") # read: csv, None; save: csv, None
    fourth = s.datasets["cpi"].copy()

    fifth = third.copy()
    fifth.columns = fifth.columns.get_level_values(0)
    fifth.rename_axis(None, axis=1, inplace=True)

    assert all(v < 1.01 and v > 0.99 for v in (second / first).values)
    assert (second.columns == first.columns)[0] and (second.index == first.index)[0]
    assert all(v < 1.01 and v > 0.99 for v in (third / second).values)
    assert (third.columns == second.columns)[0] and (third.index == second.index)[0]
    assert all(v < 1.01 and v > 0.99 for v in (fifth / fourth).values)
    assert (fifth.columns == fourth.columns)[0] and (fifth.index == fourth.index)[0]


@pytest.mark.parametrize("log_level,file", [("test", "test.log"), (2, "info.log")])
def test_logging_file(caplog, log_level, file, tmpdir):
    remove_temporary_files_folders()
    caplog.clear()
    Session(location=tmpdir, log=log_level)
    test_path = path.join(tmpdir, file)
    assert path.isfile(test_path)
    logging.shutdown()
    caplog.clear()
    remove_temporary_files_folders()


def test_logging_error(caplog):
    with pytest.raises(ValueError):
        Session(location=TEST_DIR, log=5)
    caplog.clear()


@pytest.mark.parametrize("freq_resample,freq_result", [("A-DEC", "A-DEC"), ("M", "Q-DEC")])
def test_concat(freq_resample, freq_result):
    s = Session(location=TEST_CON, download=False)
    s.get(["natacc_ind_con_nsa", "public_debt_gps"])
    s.resample(select="natacc_ind_con_nsa", rule=freq_resample)
    s.concat(select="all", concat_name="test")
    df = s.datasets["concat_test"]
    combined_cols = (s.datasets["natacc_ind_con_nsa"].columns
                     .append(s.datasets["public_debt_gps"].columns)
                     .get_level_values(0))
    assert pd.infer_freq(df.index) == freq_result
    assert all(x in df.columns.get_level_values(0) for x in combined_cols)



from os import PathLike
from pathlib import Path
from io import BytesIO
from typing import Union
from urllib.error import URLError, HTTPError

import pandas as pd
import requests
from opnieuw import retry
from pandas.tseries.offsets import MonthEnd
from sqlalchemy.engine.base import Connection, Engine

from econuy.utils import ops, metadata, get_project_root
from econuy.utils.sources import urls


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def income_household() -> pd.DataFrame:
    """Get average household income.

    Parameters
    ----------
    update_loc : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
                  default None
        Either Path or path-like string pointing to a directory where to find
        a CSV for updating, SQLAlchemy connection or engine object, or
        ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_loc : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
                default None
        Either Path or path-like string pointing to a directory where to save
        the CSV, SQL Alchemy connection or engine object, or ``None``,
        don't save.
    only_get : bool, default False
        If True, don't download data, retrieve what is available from
        ``update_loc``.

    Returns
    -------
    Monthly average household income : pd.DataFrame

    """
    name = "income_household"

    try:
        raw = pd.read_excel(urls[name]["dl"]["main"], sheet_name="Mensual",
                            skiprows=5, index_col=0).dropna(how="all")
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["main"],
                             verify=certificate)
            raw = pd.read_excel(BytesIO(r.content),
                                sheet_name="Mensual",
                                skiprows=5, index_col=0).dropna(how="all")
        else:
            raise err
    raw.index = pd.to_datetime(raw.index)
    output = raw.loc[~pd.isna(raw.index)]
    output.index = output.index + MonthEnd(0)
    output.columns = ["Total país", "Montevideo", "Interior: total",
                      "Interior: localidades de más de 5 mil hab.",
                      "Interior: localidades pequeñas y rural"]

    missing = pd.read_excel(urls[name]["dl"]["missing"],
                            index_col=0, header=0).iloc[:, 10:13]
    missing.columns = output.columns[:3]
    output = output.append(missing, sort=False)
    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Ingresos", currency="UYU",
                  inf_adj="No", unit="Pesos", seas_adj="NSA",
                  ts_type="Flujo", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def income_capita(update_loc: Union[str, PathLike,
                                    Engine, Connection, None] = None,
                  revise_rows: Union[str, int] = "nodup",
                  save_loc: Union[str, PathLike,
                                  Engine, Connection, None] = None,
                  only_get: bool = False) -> pd.DataFrame:
    """Get average per capita income.

    Parameters
    ----------
    update_loc : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
                  default None
        Either Path or path-like string pointing to a directory where to find
        a CSV for updating, SQLAlchemy connection or engine object, or
        ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_loc : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
                default None
        Either Path or path-like string pointing to a directory where to save
        the CSV, SQL Alchemy connection or engine object, or ``None``,
        don't save.
    only_get : bool, default False
        If True, don't download data, retrieve what is available from
        ``update_loc``.

    Returns
    -------
    Monthly average per capita income : pd.DataFrame

    """
    name = "income_capita"

    if only_get is True and update_loc is not None:
        output = ops._io(operation="read", data_loc=update_loc,
                         name=name)
        if not output.equals(pd.DataFrame()):
            return output
    try:
        raw = pd.read_excel(urls[name]["dl"]["main"],
                            sheet_name="Mensuall", skiprows=5,
                            index_col=0).dropna(how="all")
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["main"],
                             verify=certificate)
            raw = pd.read_excel(BytesIO(r.content),
                                sheet_name="Mensuall", skiprows=5,
                                index_col=0).dropna(how="all")
        else:
            raise err
    raw.index = pd.to_datetime(raw.index)
    output = raw.loc[~pd.isna(raw.index)]
    output.index = output.index + MonthEnd(0)
    output.columns = ["Total país", "Montevideo", "Interior: total",
                      "Interior: localidades de más de 5 mil hab.",
                      "Interior: localidades pequeñas y rural"]

    missing = pd.read_excel(urls[name]["dl"]["missing"],
                            index_col=0, header=0).iloc[:, 13:16]
    missing.columns = output.columns[:3]
    output = output.append(missing, sort=False)
    output.rename_axis(None, inplace=True)

    if update_loc is not None:
        previous_data = ops._io(operation="read",
                                data_loc=update_loc,
                                name=name)
        output = ops._revise(new_data=output, prev_data=previous_data,
                             revise_rows=revise_rows)

    metadata._set(output, area="Ingresos", currency="UYU",
                  inf_adj="No", unit="Pesos", seas_adj="NSA",
                  ts_type="Flujo", cumperiods=1)

    if save_loc is not None:
        ops._io(operation="save", data_loc=save_loc,
                data=output, name=name)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def consumer_confidence(
        update_loc: Union[str, PathLike, Engine, Connection, None] = None,
        revise_rows: Union[str, int] = "nodup",
        save_loc: Union[str, PathLike, Engine, Connection, None] = None,
        only_get: bool = False) -> pd.DataFrame:
    """Get monthly consumer confidence data.

    Parameters
    ----------
    update_loc : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
                  default None
        Either Path or path-like string pointing to a directory where to find
        a CSV for updating, SQLAlchemy connection or engine object, or
        ``None``, don't update.
    revise_rows : {'nodup', 'auto', int}
        Defines how to process data updates. An integer indicates how many rows
        to remove from the tail of the dataframe and replace with new data.
        String can either be ``auto``, which automatically determines number of
        rows to replace from the inferred data frequency, or ``nodup``,
        which replaces existing periods with new data.
    save_loc : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
                default None
        Either Path or path-like string pointing to a directory where to save
        the CSV, SQL Alchemy connection or engine object, or ``None``,
        don't save.
    only_get : bool, default False
        If True, don't download data, retrieve what is available from
        ``update_loc``.

    Returns
    -------
    Monthly consumer confidence data : pd.DataFrame

    """
    name = "consumer_confidence"

    if only_get is True and update_loc is not None:
        output = ops._io(operation="read", data_loc=update_loc,
                         name=name)
        if not output.equals(pd.DataFrame()):
            return output

    raw = pd.read_excel(urls[name]["dl"]["main"], skiprows=3,
                        usecols="B:F", index_col=0)
    output = raw.loc[~pd.isna(raw.index)]
    output.index = output.index + MonthEnd(0)
    output.columns = ["Subíndice: Situación Económica Personal",
                      "Subíndice: Situación Económica del País",
                      "Subíndice: Predisposición a la Compra de Durables",
                      "Índice de Confianza del Consumidor"]
    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)

    if update_loc is not None:
        previous_data = ops._io(operation="read", data_loc=update_loc,
                                name=name)
        output = ops._revise(new_data=output, prev_data=previous_data,
                             revise_rows=revise_rows)

    metadata._set(output, area="Actividad económica", currency="-",
                  inf_adj="No", unit="50 = neutralidad", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    if save_loc is not None:
        ops._io(operation="save", data_loc=save_loc,
                data=output, name=name)

    return output



import datetime as dt
import re
import time
from io import BytesIO
from typing import Optional
from urllib.error import HTTPError, URLError
from requests.exceptions import SSLError

import pandas as pd
import requests
from bs4 import BeautifulSoup
from opnieuw import retry
from pandas.tseries.offsets import MonthEnd
from selenium.webdriver.remote.webdriver import WebDriver

from econuy.utils import metadata
from econuy.utils.chromedriver import _build
from econuy.utils.sources import urls
from econuy.core import Pipeline


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def credit() -> pd.DataFrame:
    """Get bank credit data.

    Returns
    -------
    Monthly credit : pd.DataFrame

    """
    name = "credit"

    raw = pd.read_excel(urls[name]["dl"]["main"],
                        sheet_name="Total Sist. Banc.",
                        skiprows=10, usecols="A:P,T:AB,AD:AL", index_col=0)
    output = raw.loc[~pd.isna(raw.index)].dropna(how="all", axis=1)
    output.index = output.index + MonthEnd(0)
    output.columns = ["Créditos: Resid. privado - vigentes",
                      "Créditos: Resid. privado - vencidos",
                      "Créditos: Resid. privado- total",
                      "Créditos: Resid. público - vigentes",
                      "Créditos: Resid. público - vencidos",
                      "Créditos: Resid. público - total",
                      "Créditos: Resid. total - vigentes",
                      "Créditos: Resid. total - vencidos",
                      "Créditos: Resid. total - total",
                      "Créditos: No residentes - vigentes",
                      "Créditos: No residentes - vencidos",
                      "Créditos: No residentes - total",
                      "Créditos: Total - vigentes",
                      "Créditos: Total - vencidos",
                      "Créditos: Total - total",
                      "Créditos: Resid. MN privado - vigentes",
                      "Créditos: Resid. MN privado - vencidos",
                      "Créditos: Resid. MN privado- total",
                      "Créditos: Resid. MN público - vigentes",
                      "Créditos: Resid. MN público - vencidos",
                      "Créditos: Resid. MN público- total",
                      "Créditos: Resid. MN total - vigentes",
                      "Créditos: Resid. MN total - vencidos",
                      "Créditos: Resid. MN total- total",
                      "Créditos: Resid. ME privado - vigentes",
                      "Créditos: Resid. ME privado - vencidos",
                      "Créditos: Resid. ME privado- total",
                      "Créditos: Resid. ME público - vigentes",
                      "Créditos: Resid. ME público - vencidos",
                      "Créditos: Resid. ME público- total",
                      "Créditos: Resid. ME total - vigentes",
                      "Créditos: Resid. ME total - vencidos",
                      "Créditos: Resid. ME total- total"]

    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Sector financiero", currency="USD",
                  inf_adj="No", unit="Millones", seas_adj="NSA",
                  ts_type="Stock", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def deposits() -> pd.DataFrame:
    """Get bank deposits data.

    Returns
    -------
    Monthly deposits : pd.DataFrame

    """
    name = "deposits"

    raw = pd.read_excel(urls[name]["dl"]["main"],
                        sheet_name="Total Sist. Banc.",
                        skiprows=8, usecols="A:S", index_col=0)
    output = raw.loc[~pd.isna(raw.index)].dropna(how="all", axis=1)
    output.index = output.index + MonthEnd(0)
    output.columns = ["Depósitos: S. privado - MN",
                      "Depósitos: S. privado - ME",
                      "Depósitos: S. privado - total",
                      "Depósitos: S. público - MN",
                      "Depósitos: S. público - ME",
                      "Depósitos: S. público - total",
                      "Depósitos: Total - MN",
                      "Depósitos: Total - ME",
                      "Depósitos: Total - total",
                      "Depósitos: S. privado - MN vista",
                      "Depósitos: S. privado - MN plazo",
                      "Depósitos: S. privado - ME vista",
                      "Depósitos: S. privado - ME plazo",
                      "Depósitos: S. privado - total vista",
                      "Depósitos: S. privado - total plazo",
                      "Depósitos: S. privado - residente",
                      "Depósitos: S. privado - no residente"]

    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Sector financiero", currency="USD",
                  inf_adj="No", unit="Millones", seas_adj="NSA",
                  ts_type="Stock", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def interest_rates() -> pd.DataFrame:
    """Get interest rates data.

    Returns
    -------
    Monthly interest rates : pd.DataFrame

    """
    name = "interest_rates"

    xls = pd.ExcelFile(urls["interest_rates"]["dl"]["main"])
    sheets = ["Activas $", "Activas UI", "Activas U$S",
              "Pasivas $", "Pasivas UI", "Pasivas U$S"]
    columns = ["B:C,G,K", "B:C,G,K", "B:C,H,L",
               "B:C,N,T", "B:C,P,W", "B:C,N,T"]
    sheet_data = []
    for sheet, columns in zip(sheets, columns):
        if "Activas" in sheet:
            skip = 11
        else:
            skip = 10
        data = pd.read_excel(xls, sheet_name=sheet, skiprows=skip,
                             usecols=columns, index_col=0)
        data.index = pd.to_datetime(data.index, errors="coerce")
        data = data.loc[~pd.isna(data.index)]
        data.index = data.index + MonthEnd(0)
        sheet_data.append(data)
    output = pd.concat(sheet_data, axis=1)
    output.columns = ["Tasas activas: $, promedio",
                      "Tasas activas: $, promedio empresas",
                      "Tasas activas: $, promedio familias",
                      "Tasas activas: UI, promedio",
                      "Tasas activas: UI, promedio empresas",
                      "Tasas activas: UI, promedio familias",
                      "Tasas activas: US$, promedio",
                      "Tasas activas: US$, promedio empresas",
                      "Tasas activas: US$, promedio familias",
                      "Tasas pasivas: $, promedio",
                      "Tasas pasivas: $, promedio empresas",
                      "Tasas pasivas: $, promedio familias",
                      "Tasas pasivas: UI, promedio",
                      "Tasas pasivas: UI, promedio empresas",
                      "Tasas pasivas: UI, promedio familias",
                      "Tasas pasivas: US$, promedio",
                      "Tasas pasivas: US$, promedio empresas",
                      "Tasas pasivas: US$, promedio familias"]

    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Sector financiero", currency="-",
                  inf_adj="-", unit="Tasa", seas_adj="NSA",
                  ts_type="Flujo", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3, 4],
                                new_arrays=[["UYU", "UYU", "UYU",
                                             "UYU", "UYU", "UYU",
                                             "USD", "USD", "USD",
                                             "UYU", "UYU", "UYU",
                                             "UYU", "UYU", "UYU",
                                             "USD", "USD", "USD"],
                                            ["No", "No", "No",
                                             "Const.", "Const.", "Const.",
                                             "No", "No", "No",
                                             "No", "No", "No",
                                             "Const.", "Const.", "Const.",
                                             "No", "No", "No"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def sovereign_risk() -> pd.DataFrame:
    """Get Uruguayan Bond Index (sovereign risk spreads) data.

    Returns
    -------
    Uruguayan Bond Index : pd.DataFrame

    """
    name = "sovereign_risk"

    try:
        historical = pd.read_excel(urls[name]["dl"]["historical"],
                                   usecols="B:C", skiprows=1, index_col=0,
                                   sheet_name="Valores de Cierre Diarios")
        r_current = requests.get(urls[name]["dl"]["current"])
    except (SSLError, URLError, HTTPError):
        r_historical = requests.get(
            urls[name]["dl"]["historical"], verify=False)
        historical = pd.read_excel(BytesIO(r_historical.content),
                                   usecols="B:C", skiprows=1, index_col=0,
                                   sheet_name="Valores de Cierre Diarios")
        r_current = requests.get(urls[name]["dl"]["current"], verify=False)
    soup = BeautifulSoup(r_current.text, features="lxml")
    raw_string = soup.find_all(type="hidden")[0]["value"]
    raw_list = raw_string.split("],")
    raw_list = [re.sub(r'["\[\]]', "", line) for line in raw_list]
    index = [x.split(",")[0] for x in raw_list]
    values = [x.split(",")[1] for x in raw_list]
    current = pd.DataFrame(data=values, index=index, columns=["UBI"])
    current.index = pd.to_datetime(current.index, format="%d/%m/%y")
    output = pd.concat([historical, current])
    output = output.loc[~output.index.duplicated(keep="last")]
    output.sort_index(inplace=True)
    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Sector financiero", currency="USD",
                  inf_adj="No", unit="PBS", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def call_rate(driver: Optional[WebDriver] = None) -> pd.DataFrame:
    """Get 1-day call interest rate data.

    This function requires a Selenium webdriver. It can be provided in the
    driver parameter, or it will attempt to configure a Chrome webdriver.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver, default None
        Selenium webdriver for scraping. If None, build a Chrome webdriver.

    Returns
    -------
    Daily call rate : pd.DataFrame

    """
    name = "call"

    if driver is None:
        driver = _build()
    driver.get(urls[name]["dl"]["main"])
    start = driver.find_element(
        by="name", value="ctl00$ContentPlaceHolder1$dateDesde$dateInput")
    start.clear()
    start.send_keys("01/01/2002")
    end = driver.find_element(
        by="name",
        value="ctl00$ContentPlaceHolder1$dateHasta$dateInput")
    end.clear()
    end.send_keys(dt.datetime.now().strftime("%d/%m/%Y"))
    submit = driver.find_element(by="id",
                                 value="ContentPlaceHolder1_LinkFiltrar")
    submit.click()
    time.sleep(5)
    tables = pd.read_html(driver.page_source, decimal=",", thousands=".")
    driver.quit()
    raw = tables[8].iloc[:, :-2]
    call = raw.set_index("FECHA")
    call.index = pd.to_datetime(call.index, format="%d/%m/%Y")
    call.sort_index(inplace=True)
    call.columns = ["Tasa call a 1 día: Promedio",
                    "Tasa call a 1 día: Máximo",
                    "Tasa call a 1 día: Mínimo"]
    call = call.apply(pd.to_numeric, errors="coerce")
    call.rename_axis(None, inplace=True)

    metadata._set(call, area="Sector financiero", currency="UYU",
                  inf_adj="No", unit="Tasa", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    return call


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=8,
    retry_window_after_first_call_in_seconds=60,
)
def bonds(driver: Optional[WebDriver] = None) -> pd.DataFrame:
    """Get interest rate yield for Uruguayan US-denominated bonds,
    inflation-linked bonds and peso bonds.

    This function requires a Selenium webdriver. It can be provided in the
    driver parameter, or it will attempt to configure a Chrome webdriver.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver, default None
        Selenium webdriver for scraping. If None, build a Chrome webdriver.

    Returns
    -------
    Daily bond yields in basis points : pd.DataFrame

    """
    name = "bonds"

    if driver is None:
        driver = _build()
    dfs = []
    for url in urls[name]["dl"].values():
        driver.get(url)
        start = driver.find_element(
            by="name", value="ctl00$ContentPlaceHolder1$dateDesde$dateInput")
        start.clear()
        start.send_keys("01/01/2000")
        end = driver.find_element(
            by="name", value="ctl00$ContentPlaceHolder1$dateHasta$dateInput")
        end.clear()
        end.send_keys(dt.datetime.now().strftime("%d/%m/%Y"))
        submit = driver.find_element(by="id",
                                     value="ContentPlaceHolder1_LinkFiltrar")
        submit.click()
        time.sleep(10)
        tables = pd.read_html(driver.page_source, decimal=",", thousands=".")

        raw = tables[8]
        df = raw.set_index("FECHA")
        df.index = pd.to_datetime(df.index, format="%d/%m/%Y")
        df.sort_index(inplace=True)
        df = df.loc[:, df.columns.isin(["BPS", "RENDIMIENTO"])]
        df.columns = [url]
        dfs.append(df)
    driver.quit()
    output = dfs[0].join(dfs[1], how="outer").join(dfs[2], how="outer")
    output.columns = ["Bonos soberanos en dólares", "Bonos soberanos en UI",
                      "Bonos soberanos en pesos"]
    output = output.loc[~output.index.duplicated()]

    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Sector financiero", currency="-",
                  inf_adj="No", unit="PBS", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3, 4],
                                new_arrays=[["USD", "UYU", "UYU"],
                                            ["No", "Const.", "No"]])

    return output



from __future__ import annotations
import copy
from typing import Union, Optional, Dict
from os import PathLike
from datetime import datetime

import pandas as pd
from sqlalchemy.engine.base import Engine, Connection

from econuy.utils import ops, datasets
from econuy import transform


class Pipeline(object):
    """
    Main class to access download and transformation methods.

    Attributes
    ----------
    location : str, os.PathLike, SQLAlchemy Connection or Engine, or None, \
               default None
        Either Path or path-like string pointing to a directory where to find
        a CSV for updating and saving, SQLAlchemy connection or engine object,
        or ``None``, don't save or update.
    download : bool, default True
        If False the ``get`` method will only try to retrieve data on disk.
    always_Save : bool, default True
        If True, save every retrieved dataset to the specified ``location``.
    read_fmt : {'csv', 'xls', 'xlsx'}
        File format of previously downloaded data. Ignored if ``location``
        points to a SQL object.
    save_fmt : {'csv', 'xls', 'xlsx'}
        File format for saving. Ignored if ``location`` points to a SQL object.
    read_header : {'included', 'separate', None}
        Location of dataset metadata headers. 'included' means they are in the
        first 9 rows of the dataset. 'separate' means they are in a separate
        Excel sheet (if ``read_fmt='csv'``, headers are discarded).
        None means there are no metadata headers.
    save_header : {'included', 'separate', None}
        Location of dataset metadata headers. 'included' means they will be set
        as the first 9 rows of the dataset. 'separate' means they will be saved
        in a separate Excel sheet (if ``save_fmt='csv'``, headers are
        discarded). None discards any headers.
    errors : {'raise', 'coerce', 'ignore'}
        How to handle errors that arise from transformations. ``raise`` will
        raise a ValueError, ``coerce`` will force the data into ``np.nan`` and
        ``ignore`` will leave the input data as is.

    """
    def __init__(self,
                 location: Union[str, PathLike, Engine,
                                 Connection, None] = None,
                 download: bool = True,
                 always_save: bool = True,
                 read_fmt: str = "csv",
                 read_header: Optional[str] = "included",
                 save_fmt: str = "csv",
                 save_header: Optional[str] = "included",
                 errors: str = "raise"):
        self.location = location
        self.download = download
        self.always_save = always_save
        self.read_fmt = read_fmt
        self.read_header = read_header
        self.save_fmt = save_fmt
        self.save_header = save_header
        self.errors = errors
        self._name = None
        self._dataset = pd.DataFrame()
        self._download_commodity_weights = False

    @property
    def dataset(self) -> pd.DataFrame:
        """Get dataset."""
        return self._dataset

    @property
    def dataset_flat(self) -> pd.DataFrame:
        """Get dataset with no metadata in its column names."""
        nometa = self._dataset.copy(deep=True)
        nometa.columns = nometa.columns.get_level_values(0)
        return nometa

    @property
    def name(self) -> str:
        """Get dataset name."""
        return self._name

    @property
    def description(self) -> str:
        """Get dataset description."""
        try:
            return self.available_datasets()[self.name]["description"]
        except KeyError:
            return None

    @staticmethod
    def available_datasets() -> Dict:
        """Get a dictionary with all available datasets.

        The dictionary is separated by original and custom keys, which denote
        whether the dataset has been modified in some way or if its as provided
        by the source
        """
        all = datasets.original()
        all.update(datasets.custom())
        return all

    def __repr__(self):
        return (f"Pipeline(location={self.location})\n"
                f"Current dataset: {self.name}")

    def copy(self, deep: bool = False) -> Pipeline:
        """Copy or deepcopy a Pipeline object.

        Parameters
        ----------
        deep : bool, default True
            If True, deepcopy.

        Returns
        -------
        :class:`~econuy.core.Pipeline`

        """
        if deep:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    def get(self, name: str):
        """
        Main download method.

        Parameters
        ----------
        name : str
            Dataset to download, see available options in
            :mod:`~econuy.core.Pipeline.available_datasets`.

        Raises
        ------
        ValueError
            If an invalid string is given to the ``name`` argument.

        """
        if name not in self.available_datasets().keys():
            raise ValueError("Invalid dataset selected.")

        if self.location is None:
            prev_data = pd.DataFrame()
        else:
            prev_data = ops._io(operation="read", data_loc=self.location,
                                name=name, file_fmt=self.read_fmt,
                                multiindex=self.read_header)
        if not self.download and not prev_data.empty:
            self._dataset = prev_data
            print(f"{name}: previous data found.")
        else:
            if not self.download:
                print(f"{name}: previous data not found, downloading.")
            selection = self.available_datasets()[name]
            if name in ["trade_balance", "terms_of_trade", "rxr_custom",
                           "commodity_index", "cpi_measures", "_lin_gdp",
                           "net_public_debt", "balance_summary",
                           "core_industrial", "regional_embi_yields",
                           "regional_rxr", "regional_stocks",
                           "real_wages", "labor_rates_people"]:
                # Some datasets require retrieving other datasets. Passing the
                # class instance allows running these retrieval operations
                # with the same parameters (for example, save file formats).
                new_data = selection["function"](pipeline=self)
            elif name in ["reserves_changes", "nxr_daily"]:
                # Datasets that require many requests (mostly daily data) benefit
                # from having previous data, so they can start requests
                # from the last available date.
                new_data = selection["function"](pipeline=self,
                                                 previous_data=prev_data)
            else:
                new_data = selection["function"]()
            data = ops._revise(new_data=new_data, prev_data=prev_data,
                               revise_rows="nodup")
            self._dataset = data
        self._name = name
        if (self.always_save
            and (self.download or prev_data.empty)
            and self.location is not None):
            self.save()
        return

    def resample(self, rule: Union[pd.DateOffset, pd.Timedelta, str],
                 operation: str = "sum",  interpolation: str = "linear",
                 warn: bool = False):
        """
        Wrapper for the `resample method <https://pandas.pydata.org/pandas-docs
        stable/reference/api/pandas.DataFrame.resample.html>`_ in Pandas that
        integrates with econuy dataframes' metadata.

        Trim partial bins, i.e. do not calculate the resampled
        period if it is not complete, unless the input dataframe has no defined
        frequency, in which case no trimming is done.

        Parameters
        ----------
        rule : pd.DateOffset, pd.Timedelta or str
            Target frequency to resample to. See
            `Pandas offset aliases <https://pandas.pydata.org/pandas-docs/stable/
            user_guide/timeseries.html#offset-aliases>`_
        operation : {'sum', 'mean', 'last', 'upsample'}
            Operation to use for resampling.
        interpolation : str, default 'linear'
            Method to use when missing data are produced as a result of
            resampling, for example when upsampling to a higher frequency. See
            `Pandas interpolation methods <https://pandas.pydata.org/pandas-docs
            /stable/reference/api/pandas.Series.interpolate.html>`_
        warn : bool, default False
            If False, don't raise warnings with incomplete time-range bins.

        Returns
        -------
        ``None``

        Raises
        ------
        ValueError
            If ``operation`` is not one of available options.
        ValueError
            If the input dataframe's columns do not have the appropiate levels.

        Warns
        -----
        UserWarning
            If input frequencies cannot be assigned a numeric value, preventing
            incomplete bin trimming.

        """
        if self.dataset.empty:
            raise ValueError("Can't use transformation methods without "
                             "retrieving a dataset first.")
        output = transform.resample(self.dataset, rule=rule,
                                    operation=operation,
                                    interpolation=interpolation, warn=warn)
        self._dataset = output
        return

    def chg_diff(self, operation: str = "chg",
                 period: str = "last"):
        """Wrapper for the `pct_change <https://pandas.pydata.org/pandas-docs/stable/
        reference/api/pandas.DataFrame.pct_change.html>`_ and `diff <https://pandas
        .pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.diff.html>`_
        Pandas methods.

        Calculate percentage change or difference for dataframes. The ``period``
        argument takes into account the frequency of the dataframe, i.e.,
        ``inter`` (for interannual) will calculate pct change/differences with
        ``periods=4`` for quarterly frequency, but ``periods=12`` for monthly
        frequency.

        Parameters
        ----------
        df : pd.DataFrame
            Input dataframe.
        operation : {'chg', 'diff'}
            ``chg`` for percent change or ``diff`` for differences.
        period : {'last', 'inter', 'annual'}
            Period with which to calculate change or difference. ``last`` for
            previous period (last month for monthly data), ``inter`` for same
            period last year, ``annual`` for same period last year but taking
            annual sums.

        Returns
        -------
        ``None``

        Raises
        ------
        ValueError
            If the dataframe is not of frequency ``M`` (month), ``Q`` or
            ``Q-DEC`` (quarter), or ``A`` or ``A-DEC`` (year).
        ValueError
            If the ``operation`` parameter does not have a valid argument.
        ValueError
            If the ``period`` parameter does not have a valid argument.
        ValueError
            If the input dataframe's columns do not have the appropiate levels.

        """
        if self.dataset.empty:
            raise ValueError("Can't use transformation methods without "
                             "retrieving a dataset first.")
        output = transform.chg_diff(
            self.dataset, operation=operation, period=period)
        self._dataset = output
        return

    def decompose(self, component: str = "both",
                  method: str = "x13",
                  force_x13: bool = False,
                  fallback: str = "loess",
                  trading: bool = True,
                  outlier: bool = True,
                  x13_binary: Union[str, PathLike] = "search",
                  search_parents: int = 1,
                  ignore_warnings: bool = True,
                  **kwargs):
        """Apply seasonal decomposition.

        Decompose the series in a Pandas dataframe using either X13 ARIMA, Loess
        or moving averages. X13 can be forced in case of failure by alternating
        the underlying function's parameters. If not, it will fall back to one of
        the other methods. If the X13 method is chosen, the X13 binary has to be
        provided. Please refer to the README for instructions on where to get this
        binary.

        Parameters
        ----------
        component : {'seas', 'trend'}
            Return seasonally adjusted or trend component.
        method : {'x13', 'loess', 'ma'}
            Decomposition method. ``X13`` refers to X13 ARIMA from the US Census,
            ``loess`` refers to Loess decomposition and ``ma`` refers to moving
            average decomposition, in all cases as implemented by
            `statsmodels <https://www.statsmodels.org/dev/tsa.html>`_.
        force_x13 : bool, default False
            Whether to try different ``outlier`` and ``trading`` parameters
            in statsmodels' `x13 arima analysis <https://www.statsmodels.org/dev/
            generated/statsmodels.tsa.x13.x13_arima_analysis.html>`_ for each
            series that fails. If ``False``, jump to the ``fallback`` method for
            the whole dataframe at the first error.
        fallback : {'loess', 'ma'}
            Decomposition method to fall back to if ``method="x13"`` fails and
            ``force_x13=False``.
        trading : bool, default True
            Whether to automatically detect trading days in X13 ARIMA.
        outlier : bool, default True
            Whether to automatically detect outliers in X13 ARIMA.
        x13_binary: str, os.PathLike or None, default 'search'
            Location of the X13 binary. If ``search`` is used, will attempt to find
            the binary in the project structure. If ``None``, statsmodels will
            handle it.
        search_parents: int, default 1
            If ``x13_binary=search``, this parameter controls how many parent
            directories to go up before recursively searching for the binary.
        ignore_warnings : bool, default True
            Whether to suppress X13Warnings from statsmodels.
        kwargs
            Keyword arguments passed to statsmodels' ``x13_arima_analysis``,
            ``STL`` and ``seasonal_decompose``.

        Returns
        -------
        ``None``

        Raises
        ------
        ValueError
            If the ``method`` parameter does not have a valid argument.
        ValueError
            If the ``component`` parameter does not have a valid argument.
        ValueError
            If the ``fallback`` parameter does not have a valid argument.
        ValueError
            If the ``errors`` parameter does not have a valid argument.
        FileNotFoundError
            If the path provided for the X13 binary does not point to a file and
            ``method='x13'``.

        """
        if self.dataset.empty:
            raise ValueError("Can't use transformation methods without "
                             "retrieving a dataset first.")
        valid_component = ["seas", "trend"]
        if component not in valid_component:
            raise ValueError(f"Only {', '.join(valid_component)} are allowed."
                             f"See underlying 'decompose'.")

        output = transform.decompose(self.dataset, component=component,
                                     method=method,
                                     force_x13=force_x13,
                                     fallback=fallback,
                                     trading=trading,
                                     outlier=outlier,
                                     x13_binary=x13_binary,
                                     search_parents=search_parents,
                                     ignore_warnings=ignore_warnings,
                                     errors=self.errors,
                                     **kwargs)
        self._dataset = output
        return

    def convert(self, flavor: str,
                start_date: Union[str, datetime, None] = None,
                end_date: Union[str, datetime, None] = None):
        """Convert dataframe from UYU to USD, from UYU to real UYU or
        from UYU/USD to % GDP.

        ``flavor=usd``: Convert a dataframe's columns from Uruguayan pesos to US dollars. Call the
        :mod:`~econuy.core.Pipeline.get` function to obtain nominal
        exchange rates, and take into account whether the input dataframe's
        ``Type``, as defined by its multiindex, is flow or stock, in order to `
        choose end of period or monthly average NXR. Also take into account the
        input dataframe's frequency and whether columns represent rolling averages
        or sums.

        ``flavor=real``: Convert a dataframe's columns to real prices. Call the
        :mod:`~econuy.core.Pipeline.get` method to obtain the consumer price
        index. take into account the input dataframe's frequency and whether
        columns represent rolling averages or sums. Allow choosing a single period,
        a range of dates or no period as a base (i.e., period for which the
        average/sum of input dataframe and output dataframe is the same).

        ``flavor=gdp``: Convert a dataframe's columns to percentage of GDP. Call the
        the :mod:`~econuy.core.Pipeline.get` method to obtain UYU and USD quarterly GDP series.
        Take into account the input dataframe's
        currency for chossing UYU or USD GDP. If frequency of input dataframe is
        higher than quarterly, GDP will be upsampled and linear interpolation will
        be performed to complete missing data.
        If input dataframe's "Acum." level is not 12 for monthly frequency or 4
        for quarterly frequency, calculate rolling input dataframe.

        In all cases, if input dataframe's frequency is higher than monthly
        (daily, business, etc.), resample to monthly frequency.

        Parameters
        ----------
        pipeline : econuy.core.Pipeline or None, default None
            An instance of the econuy Pipeline class.
        start_date : str, datetime.date or None, default None
            Only used if ``flavor=real``. If set to a date-like string or a
            date, and ``end_date`` is None, the base period will be
            ``start_date``.
        end_date : str, datetime.date or None, default None
            Only used if ``flavor=real``. If ``start_date`` is set, calculate
            so that the data is in constant prices of ``start_date-end_date``.
        errors : {'raise', 'coerce', 'ignore'}
            What to do when a column in the input dataframe is not expressed in
            Uruguayan pesos. ``raise`` will raise a ValueError, ``coerce`` will
            force the entire column into ``np.nan`` and ``ignore`` will leave the
            input column as is.

        Returns
        -------
        ``None``

        Raises
        ------
        ValueError
            If the ``errors`` parameter does not have a valid argument.
        ValueError
            If the input dataframe's columns do not have the appropiate levels.

        """
        if self.dataset.empty:
            raise ValueError("Can't use transformation methods without "
                             "retrieving a dataset first.")
        if flavor not in ["usd", "real", "gdp", "pcgdp"]:
            raise ValueError("'flavor' can be one of 'usd', 'real', "
                             "or 'gdp'.")

        if flavor == "usd":
            output = transform.convert_usd(self.dataset,
                                           errors=self.errors,
                                           pipeline=self)
        elif flavor == "real":
            output = transform.convert_real(self.dataset, start_date=start_date,
                                            end_date=end_date,
                                            errors=self.errors,
                                            pipeline=self)
        else:
            output = transform.convert_gdp(self.dataset, errors=self.errors,
                                           pipeline=self)

        self._dataset = output
        return

    def rebase(self, start_date: Union[str, datetime],
               end_date: Union[str, datetime, None] = None,
               base: Union[float, int] = 100.0):
        """Rebase all dataframe columns to a date or range of dates.

        Parameters
        ----------
        start_date : string or datetime.datetime
            Date to which series will be rebased.
        end_date : string or datetime.datetime, default None
            If specified, series will be rebased to the average between
            ``start_date`` and ``end_date``.
        base : float, default 100
            Float for which ``start_date`` == ``base`` or average between
            ``start_date`` and ``end_date`` == ``base``.

        Returns
        -------
        ``None``

        """
        if self.dataset.empty:
            raise ValueError("Can't use transformation methods without "
                             "retrieving a dataset first.")
        output = transform.rebase(self.dataset, start_date=start_date,
                                  end_date=end_date, base=base)
        self._dataset = output
        return

    def rolling(self, window: Optional[int] = None,
                operation: str = "sum"):
        """
        Wrapper for the `rolling method <https://pandas.pydata.org/pandas-docs/
        stable/reference/api/pandas.DataFrame.rolling.html>`_ in Pandas that
        integrates with econuy dataframes' metadata.

        If ``periods`` is ``None``, try to infer the frequency and set ``periods``
        according to the following logic: ``{'A': 1, 'Q-DEC': 4, 'M': 12}``, that
        is, each period will be calculated as the sum or mean of the last year.

        Parameters
        ----------
        window : int, default None
            How many periods the window should cover.
        operation : {'sum', 'mean'}
            Operation used to calculate rolling windows.

        Returns
        -------
        ``None``

        Raises
        ------
        ValueError
            If ``operation`` is not one of available options.
        ValueError
            If the input dataframe's columns do not have the appropiate levels.

        Warns
        -----
        UserWarning
            If the input dataframe is a stock time series, for which rolling
            operations are not recommended.

        """
        if self.dataset.empty:
            raise ValueError("Can't use transformation methods without "
                             "retrieving a dataset first.")
        output = transform.rolling(self.dataset, window=window,
                                   operation=operation)
        self._dataset = output
        return

    def save(self):
        """Write held dataset.

        Raises
        -------
        ValueError
            If `dataset` is an empty DataFrame or `self.location` is None.

        """
        if self.dataset.empty:
            raise ValueError("Can't save without "
                             "retrieving a dataset first.")
        if self.location is None:
            raise ValueError("No save location defined.")
        else:
            ops._io(operation="save", data_loc=self.location,
                    name=self.name, file_fmt=self.save_fmt,
                    multiindex=self.save_header, data=self.dataset)
        return



from os import path

import pytest
from sqlalchemy import create_engine

from econuy.utils.metadata import _get_sources
from econuy.retrieval import prices
from econuy.utils import sqlutil
try:
    from tests.test_session import remove_temporary_files_folders
except ImportError:
    from .test_session import remove_temporary_files_folders


CUR_DIR = path.abspath(path.dirname(__file__))
TEST_DIR = path.join(CUR_DIR, "test-data")
TEST_CON = create_engine("sqlite://").connect()
sqlutil.insert_csvs(con=TEST_CON, directory=TEST_DIR)


def test_sqlutil():
    remove_temporary_files_folders()
    sqlutil.read(con=TEST_CON, command='SELECT * FROM nxr_daily')
    sqlutil.read(con=TEST_CON, table_name="nxr_daily",
                 start_date="2011-01-14",
                 end_date="2012-01-15")
    sqlutil.read(con=TEST_CON, table_name="nxr_daily",
                 start_date="2011-01-14")
    sqlutil.read(con=TEST_CON, table_name="reserves_changes",
                 end_date="2012-01-01",
                 cols=["1. Compras netas de moneda extranjera"])


def test_sources():
    source_1 = _get_sources("balance_gps")
    source_2 = _get_sources("balance_nfps")
    assert source_1 == source_2



class RetryLimitError(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return self.message
        else:
            return ""



import datetime as dt
import os
import tempfile
import zipfile
from random import randint
from io import BytesIO
from os import path
from pathlib import Path
from urllib.error import HTTPError, URLError

import pandas as pd
import requests
from dotenv import load_dotenv
from opnieuw import retry
from pandas.tseries.offsets import MonthEnd

from econuy.transform import decompose, rebase
from econuy.utils import metadata, get_project_root
from econuy.utils.sources import urls
from econuy.utils.extras import investing_headers


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def gdp() -> pd.DataFrame:
    """Get seasonally adjusted real quarterly GDP for select countries.

    Countries/aggregates are US, EU-27, Japan and China.

    Returns
    -------
    Quarterly real GDP in seasonally adjusted terms : pd.DataFrame

    """
    name = "global_gdp"

    chn_y = dt.datetime.now().year + 1
    chn_r = requests.get(f"{urls[name]['dl']['chn_oecd']}{chn_y}-Q4")
    chn_json = chn_r.json()
    chn_datasets = []
    for dataset, start in zip(["0", "1"], ["2011-03-31", "1993-03-31"]):
        raw = chn_json["dataSets"][0]["series"][f"0:0:{dataset}:0"][
            "observations"]
        values = [x[0] for x in raw.values()]
        df = pd.DataFrame(data=values,
                          index=pd.date_range(start=start, freq="Q-DEC",
                                              periods=len(values)),
                          columns=["China"])
        chn_datasets.append(df)
    chn_qoq = chn_datasets[0]
    chn_yoy = chn_datasets[1]
    chn_obs = pd.read_excel(
        urls["global_gdp"]["dl"]["chn_obs"],
        index_col=0).dropna(how="all", axis=1).dropna(how="all", axis=0
                                                      )
    chn_obs = chn_obs.loc[(chn_obs.index > "2011-01-01")
                          & (chn_obs.index < "2016-01-01")]
    chn_yoy["volume"] = chn_obs
    for row in reversed(range(len(chn_yoy.loc[chn_yoy.index < "2011-01-01"]))):
        if pd.isna(chn_yoy.iloc[row, 1]):
            chn_yoy.iloc[row, 1] = (chn_yoy.iloc[row + 4, 1]
                                    / (1 + chn_yoy.iloc[row + 4, 0] / 100))
    chn_yoy = chn_yoy[["volume"]].loc[chn_yoy.index < "2016-01-01"]
    metadata._set(chn_yoy)
    chn_sa = decompose(chn_yoy[["volume"]].loc[chn_yoy.index < "2016-01-01"],
                       component="seas", method="x13")
    chn_sa = pd.concat([chn_sa, chn_qoq], axis=1)
    for row in range(len(chn_sa)):
        if not pd.isna(chn_sa.iloc[row, 1]):
            chn_sa.iloc[row, 0] = (chn_sa.iloc[row - 1, 0]
                                   * (1 + chn_sa.iloc[row, 1] / 100))
    chn = chn_sa.iloc[:, [0]].div(10)

    gdps = []
    load_dotenv(Path(get_project_root(), ".env"))
    fred_api_key = os.environ.get("FRED_API_KEY")
    for series in ["GDPC1", "CLVMNACSCAB1GQEU272020", "JPNRGDPEXP"]:
        r = requests.get(f"{urls[name]['dl']['fred']}{series}&api_key="
                         f"{fred_api_key}&file_type=json")
        aux = pd.DataFrame.from_records(r.json()["observations"])
        aux = aux[["date", "value"]].set_index("date")
        aux.index = pd.to_datetime(aux.index)
        aux.index = aux.index.shift(3, freq="M") + MonthEnd(0)
        aux.columns = [series]
        aux = aux.apply(pd.to_numeric, errors="coerce")
        if series == "GDPC1":
            aux = aux.div(4)
        elif series == "CLVMNACSCAB1GQEU272020":
            aux = aux.div(1000)
        gdps.append(aux)
    gdps = pd.concat(gdps, axis=1)

    output = pd.concat([gdps, chn], axis=1)
    output.columns = ["Estados Unidos", "Unión Europea", "Japón", "China"]
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Global", currency="USD",
                  inf_adj="Const.", unit="Miles de millones", seas_adj="SA",
                  ts_type="Flujo", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["USD", "EUR", "JPY", "CNY"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def stocks() -> pd.DataFrame:
    """Get stock market index data.

    Indexes selected are S&P 500, Euronext 100, Nikkei 225 and Shanghai
    Composite.

    Returns
    -------
    Daily stock market index in USD : pd.DataFrame

    """
    name = "global_stocks"

    yahoo = []
    for series in ["spy", "n100", "nikkei", "sse"]:
        aux = pd.read_csv(urls[name]["dl"][series],
                          index_col=0, usecols=[0, 4], parse_dates=True)
        aux.columns = [series]
        yahoo.append(aux)
    output = pd.concat(yahoo, axis=1).interpolate(method="linear",
                                                  limit_area="inside")
    output.columns = ["S&P 500", "Euronext 100", "Nikkei 225",
                      "Shanghai Stock Exchange Composite"]
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Global", currency="USD",
                  inf_adj="No", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["USD", "EUR", "JPY", "CNY"]])
    output = rebase(output, start_date="2019-01-02")

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def policy_rates() -> pd.DataFrame:
    """Get central bank policy interest rates data.

    Countries/aggregates selected are US, Euro Area, Japan and China.

    Returns
    -------
    Daily policy interest rates : pd.DataFrame

    """
    name = "global_policy_rates"

    r = requests.get(urls[name]["dl"]["main"])
    temp_dir = tempfile.TemporaryDirectory()
    with zipfile.ZipFile(BytesIO(r.content), "r") as f:
        f.extractall(path=temp_dir.name)
        path_temp = path.join(temp_dir.name,
                              "WEBSTATS_CBPOL_D_DATAFLOW_csv_row.csv")
        raw = pd.read_csv(path_temp, usecols=[0, 7, 19, 36, 37], index_col=0,
                          header=2, parse_dates=True).dropna(how="all")
    output = (raw.apply(pd.to_numeric, errors="coerce")
              .interpolate(method="linear", limit_area="inside"))
    output.columns = ["China", "Japón", "Estados Unidos", "Eurozona"]
    output = output[["Estados Unidos", "Eurozona", "Japón", "China"]]
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Global", currency="USD",
                  inf_adj="No", seas_adj="NSA", unit="Tasa",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["USD", "EUR", "JPY", "CNY"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def long_rates() -> pd.DataFrame:
    """Get 10-year government bonds interest rates.

    Countries/aggregates selected are US, Germany, France, Italy, Spain
    United Kingdom, Japan and China.

    Returns
    -------
    Daily 10-year government bonds interest rates : pd.DataFrame

    """
    name = "global_long_rates"

    bonds = []
    load_dotenv(Path(get_project_root(), ".env"))
    fred_api_key = os.environ.get("FRED_API_KEY")
    r = requests.get(f"{urls[name]['dl']['fred']}DGS10&api_key="
                     f"{fred_api_key}&file_type=json")
    us = pd.DataFrame.from_records(r.json()["observations"])
    us = us[["date", "value"]].set_index("date")
    us.index = pd.to_datetime(us.index)
    us.columns = ["United States"]
    bonds.append(us.apply(pd.to_numeric, errors="coerce").dropna())

    for country, sid in zip(["Germany", "France", "Italy", "Spain",
                             "United Kingdom", "Japan", "China"],
                            ["23693", "23778", "23738", "23806",
                             "23673", "23901", "29227"]):
        end_date_dt = dt.datetime(2000, 1, 1)
        start_date_dt = dt.datetime(2000, 1, 1)
        aux = []
        while end_date_dt < dt.datetime.now():
            end_date_dt = start_date_dt + dt.timedelta(days=5000)
            params = {
                "curr_id": sid,
                "smlID": str(randint(1000000, 99999999)),
                "header": f"{country} 10-Year Bond Yield Historical Data",
                "st_date": start_date_dt.strftime("%m/%d/%Y"),
                "end_date": end_date_dt.strftime("%m/%d/%Y"),
                "interval_sec": "Daily",
                "sort_col": "date",
                "sort_ord": "DESC",
                "action": "historical_data"
            }
            r = requests.post(urls["global_long_rates"]["dl"]["main"],
                              headers=investing_headers, data=params)
            aux.append(pd.read_html(r.content, match="Price",
                                    index_col=0, parse_dates=True)[0])
            start_date_dt = end_date_dt + dt.timedelta(days=1)
        aux = pd.concat(aux, axis=0)[["Price"]].sort_index()
        aux.columns = [country]
        bonds.append(aux)

    output = bonds[0].join(bonds[1:], how="left")
    output = output.interpolate(method="linear", limit_area="inside")
    output.columns = ["Estados Unidos", "Alemania", "Francia", "Italia",
                      "España", "Reino Unido", "Japón", "China"]
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Global", currency="USD",
                  inf_adj="No", seas_adj="NSA", unit="Tasa",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["USD", "EUR", "EUR", "EUR",
                                             "EUR", "GBP", "JPY", "CNY"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def nxr() -> pd.DataFrame:
    """Get currencies data.

    Selected currencies are the US dollar index, USDEUR, USDJPY and USDCNY.

    Returns
    -------
    Daily currencies : pd.DataFrame

    """
    name = "global_nxr"

    output = []
    for series in ["dollar", "eur", "jpy", "cny"]:
        aux = pd.read_csv(urls[name]["dl"][series],
                          index_col=0, usecols=[0, 4], parse_dates=True)
        aux.columns = [series]
        if series == "dollar":
            aux.dropna(inplace=True)
        output.append(aux)
    output = output[0].join(output[1:]).interpolate(method="linear",
                                                    limit_area="inside")
    output.columns = ["Índice Dólar", "Euro", "Yen", "Renminbi"]
    output.rename_axis(None, inplace=True)
    
    metadata._set(output, area="Global", currency="USD",
                  inf_adj="No", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3, 5],
                                new_arrays=[["USD", "EUR", "JPY", "CNY"],
                                            ["Canasta/USD", "EUR/USD", "JPY/USD",
                                             "CNY/USD"]])

    return output



from pathlib import Path
from io import BytesIO
from typing import Optional
from urllib.error import URLError, HTTPError

import pandas as pd
import requests
from opnieuw import retry
from pandas.tseries.offsets import MonthEnd

from econuy import transform
from econuy.core import Pipeline
from econuy.utils import metadata, get_project_root
from econuy.utils.sources import urls


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def labor_rates() -> pd.DataFrame:
    """Get labor market data (LFPR, employment and unemployment).

    Returns
    -------
    Monthly participation, employment and unemployment rates : pd.DataFrame

    """
    name = "labor_rates"

    try:
        labor_raw = pd.read_excel(
            urls[name]["dl"]["main"], skiprows=39).dropna(axis=0, thresh=2)
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["main"],
                             verify=certificate)
            labor_raw = pd.read_excel(BytesIO(r.content),
                                      skiprows=39).dropna(axis=0, thresh=2)
        else:
            raise err
    labor = labor_raw[~labor_raw["Unnamed: 0"].str.contains("-|/|Total",
                                                            regex=True)]
    labor.index = pd.date_range(start="2006-01-01",
                                periods=len(labor), freq="M")
    labor = labor.drop(columns="Unnamed: 0")
    labor.columns = ["Tasa de actividad: total", "Tasa de actividad: hombres",
                     "Tasa de actividad: mujeres", "Tasa de empleo: total",
                     "Tasa de empleo: hombres", "Tasa de empleo: mujeres",
                     "Tasa de desempleo: total", "Tasa de desempleo: hombres",
                     "Tasa de desempleo: mujeres"]
    missing = pd.read_excel(urls[name]["dl"]["missing"],
                            index_col=0, header=0).iloc[:, :9]
    missing.columns = labor.columns
    labor = labor.append(missing)
    labor = labor.loc[~labor.index.duplicated(keep="first")]
    labor = labor.apply(pd.to_numeric, errors="coerce")
    labor.rename_axis(None, inplace=True)

    metadata._set(labor, area="Mercado laboral", currency="-",
                  inf_adj="No", unit="Tasa", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    return labor


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def nominal_wages() -> pd.DataFrame:
    """Get nominal general, public and private sector wages data

    Returns
    -------
    Monthly wages separated by public and private sector : pd.DataFrame

    """
    name = "nominal_wages"

    try:
        historical = pd.read_excel(urls[name]["dl"]["historical"],
                                   skiprows=8, usecols="A:B")
        current = pd.read_excel(urls[name]["dl"]["current"],
                                skiprows=8, usecols="A,C:D")
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["historical"],
                             verify=certificate)
            historical = pd.read_excel(BytesIO(r.content),
                                       skiprows=8, usecols="A:B")
            r = requests.get(urls[name]["dl"]["current"],
                             verify=certificate)
            current = pd.read_excel(BytesIO(r.content),
                                    skiprows=8, usecols="A,C:D")
        else:
            raise err

    historical = historical.dropna(how="any").set_index("Unnamed: 0")
    current = current.dropna(how="any").set_index("Unnamed: 0")
    wages = pd.concat([historical, current], axis=1)
    wages.index = wages.index + MonthEnd(1)
    wages.columns = ["Índice medio de salarios",
                     "Índice medio de salarios privados",
                     "Índice medio de salarios públicos"]
    wages = wages.apply(pd.to_numeric, errors="coerce")
    wages.rename_axis(None, inplace=True)

    metadata._set(wages, area="Mercado laboral", currency="UYU",
                  inf_adj="No", unit="2008-07=100", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    return wages


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def hours() -> pd.DataFrame:
    """Get average hours worked data.

    Returns
    -------
    Monthly hours worked : pd.DataFrame

    """
    name = "hours_worked"

    try:
        raw = pd.read_excel(urls[name]["dl"]["main"], sheet_name="Mensual",
                            skiprows=5, index_col=0).dropna(how="all")
        prev_hours = pd.read_excel(urls[name]["dl"]["historical"], index_col=0,
                                   skiprows=8).dropna(how="all").iloc[:, [0]]
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["main"],
                             verify=certificate)
            raw = pd.read_excel(BytesIO(r.content), sheet_name="Mensual",
                                skiprows=5, index_col=0).dropna(how="all")
            r = requests.get(urls[name]["dl"]["historical"],
                             verify=certificate)
            prev_hours = pd.read_excel(BytesIO(r.content), index_col=0,
                                       skiprows=8).dropna(how="all").iloc[:, [0]]
        else:
            raise err
    raw.index = pd.to_datetime(raw.index)
    output = raw.loc[~pd.isna(raw.index)]
    output.index = output.index + MonthEnd(0)
    output.columns = ["Total", "Industrias manufactureras",
                      "Electricidad, gas, agua y saneamiento", "Construcción",
                      "Comercio", "Transporte y almacenamiento",
                      "Alojamiento y servicios de comidas",
                      "Información y comunicación", "Actividades financieras",
                      "Actividades inmobiliarias y administrativas",
                      "Actividades profesionales",
                      "Administración pública y seguridad social",
                      "Enseñanza", "Salud", "Arte y otros servicios",
                      "Act. de hogares como empleadores",
                      "Agro, forestación, pesca y minería"]

    prev_hours = prev_hours.loc[~prev_hours.index.str.contains("-|Total")]
    prev_hours.index = pd.date_range(start="2006-01-31", freq="M",
                                     periods=len(prev_hours))
    prev_hours = prev_hours.loc[prev_hours.index < "2011-01-01"]
    prev_hours.columns = ["Total"]
    output = prev_hours.append(output, sort=False)
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Mercado laboral", currency="-",
                  inf_adj="No", unit="Horas por semana", seas_adj="NSA",
                  ts_type="Stock", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def rates_people(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """
    Get labor data, both rates and persons. Extends national data between 1991
    and 2005 with data for jurisdictions with more than 5,000 inhabitants.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Labor market data : pd.DataFrame

    """
    name = "labor_rates_people"
    if pipeline is None:
        pipeline = Pipeline()

    pipeline.get("labor_rates")
    rates = pipeline.dataset
    rates = rates.loc[:, ["Tasa de actividad: total", "Tasa de empleo: total",
                          "Tasa de desempleo: total"]]
    try:
        act_5000 = pd.read_excel(urls[name]["dl"]["act_5000"],
                                 sheet_name="Mensual", index_col=0,
                                 skiprows=8, usecols="A:B").dropna(how="any")
        emp_5000 = pd.read_excel(urls[name]["dl"]["emp_5000"],
                                 sheet_name="Mensual", index_col=0,
                                 skiprows=8, usecols="A:B").dropna(how="any")
        des_5000 = pd.read_excel(urls[name]["dl"]["des_5000"],
                                 sheet_name="Mensual", index_col=0,
                                 skiprows=7, usecols="A:B").dropna(how="any")
        working_age = pd.read_excel(urls[name]["dl"]["population"],
                                    skiprows=7,
                                    index_col=0, nrows=92).dropna(how="all")
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["act_5000"],
                             verify=certificate)
            act_5000 = pd.read_excel(BytesIO(r.content), sheet_name="Mensual",
                                     index_col=0, skiprows=8,
                                     usecols="A:B").dropna(how="any")
            r = requests.get(urls[name]["dl"]["emp_5000"],
                             verify=certificate)
            emp_5000 = pd.read_excel(BytesIO(r.content), sheet_name="Mensual",
                                     index_col=0, skiprows=8,
                                     usecols="A:B").dropna(how="any")
            r = requests.get(urls[name]["dl"]["des_5000"],
                             verify=certificate)
            des_5000 = pd.read_excel(BytesIO(r.content), sheet_name="Mensual",
                                     index_col=0, skiprows=7,
                                     usecols="A:B").dropna(how="any")
            r = requests.get(urls[name]["dl"]["population"],
                             verify=certificate)
            working_age = pd.read_excel(BytesIO(r.content),
                                        skiprows=7, index_col=0,
                                        nrows=92).dropna(how="all")
        else:
            raise err
    for df in [act_5000, emp_5000, des_5000]:
        df.index = df.index + MonthEnd(0)
    rates_5000 = pd.concat([act_5000, emp_5000, des_5000], axis=1)
    rates_prev = rates_5000.loc[rates_5000.index < "2006-01-31"]
    rates_prev.columns = rates.columns
    rates = pd.concat([rates_prev, rates])
    rates.columns = rates.columns.set_levels(
        rates.columns.levels[0].str.replace(": total", ""), level=0
        )

    ages = list(range(14, 90)) + ["90 y más"]
    working_age = working_age.loc[ages].sum()
    working_age.index = pd.date_range(start="1996-06-30", end="2050-06-30",
                                      freq="A-JUN")
    monthly_working_age = working_age.resample("M").interpolate("linear")
    monthly_working_age = monthly_working_age.reindex(rates.index)
    persons = rates.iloc[:, [0, 1]].div(100).mul(monthly_working_age, axis=0)
    persons["Desempleados"] = rates.iloc[:, 2].div(100).mul(persons.iloc[:, 0])
    persons.columns = ["Activos", "Empleados", "Desempleados"]

    metadata._set(persons, area="Mercado laboral", currency="-",
                  inf_adj="No", unit="Personas", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    output = pd.concat([rates, persons], axis=1)
    output.rename_axis(None, inplace=True)

    return output


def real_wages(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """
    Get real wages.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Real wages data : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()

    pipeline.get("nominal_wages")
    wages = pipeline.dataset
    wages.columns = ["Índice medio de salarios reales",
                     "Índice medio de salarios reales privados",
                     "Índice medio de salarios reales públicos"]
    metadata._set(wages, area="Mercado laboral", currency="UYU",
                  inf_adj="Sí", seas_adj="NSA", ts_type="-", cumperiods=1)
    output = transform.convert_real(wages, pipeline=pipeline)
    output = transform.rebase(output, start_date="2008-07-31")
    output.rename_axis(None, inplace=True)

    return output



import datetime as dt
import warnings
from urllib.error import URLError, HTTPError
from io import BytesIO
from pathlib import Path
from zipfile import BadZipFile
from typing import Optional

import numpy as np
import pandas as pd
import requests
from opnieuw import retry
from pandas.tseries.offsets import MonthEnd
from scipy.stats import stats
from scipy.stats.mstats_basic import winsorize

from econuy import transform
from econuy.core import Pipeline
from econuy.utils import metadata, get_project_root
from econuy.utils.sources import urls
from econuy.utils.extras import cpi_details


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def cpi() -> pd.DataFrame:
    """Get CPI data.

    Returns
    -------
    Monthly CPI index : pd.DataFrame

    """
    name = "cpi"

    try:
        cpi = pd.read_excel(urls[name]["dl"]["main"],
                            skiprows=7, usecols="A:B",
                            index_col=0).dropna()
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["main"],
                             verify=certificate)
            cpi = pd.read_excel(BytesIO(r.content),
                                skiprows=7, usecols="A:B",
                                index_col=0).dropna()
        else:
            raise err
    cpi.columns = ["Índice de precios al consumo"]
    cpi.rename_axis(None, inplace=True)
    cpi.index = cpi.index + MonthEnd(1)
    cpi = cpi.apply(pd.to_numeric, errors="coerce")
    cpi.rename_axis(None, inplace=True)

    metadata._set(cpi, area="Precios", currency="-",
                  inf_adj="No", unit="2010-10=100", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    return cpi


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=30,
)
def nxr_monthly() -> pd.DataFrame:
    """Get monthly nominal exchange rate data.

    Returns
    -------
    Monthly nominal exchange rates : pd.DataFrame
        Sell rate, monthly average and end of period.

    """
    name = "nxr_monthly"

    try:
        nxr_raw = pd.read_excel(urls[name]["dl"]["main"], skiprows=4,
                                index_col=0, usecols="A,C,F")
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["main"],
                             verify=certificate)
            nxr_raw = pd.read_excel(BytesIO(r.content),
                                    skiprows=4, index_col=0, usecols="A,C,F")
        else:
            raise err
    nxr = nxr_raw.dropna(how="any", axis=0)
    nxr.columns = ["Tipo de cambio venta, fin de período",
                   "Tipo de cambio venta, promedio"]
    nxr.index = nxr.index + MonthEnd(1)
    nxr = nxr.apply(pd.to_numeric, errors="coerce")
    nxr.rename_axis(None, inplace=True)

    metadata._set(nxr, area="Precios", currency="UYU/USD",
                  inf_adj="No", unit="-", seas_adj="NSA",
                  ts_type="-", cumperiods=1)

    return nxr


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=10,
    retry_window_after_first_call_in_seconds=60,
)
def nxr_daily(pipeline: Optional[Pipeline] = None,
              previous_data: pd.DataFrame = pd.DataFrame()) -> pd.DataFrame:
    """Get daily nominal exchange rate data.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.
    previous_data : pd.DataFrame
        A DataFrame representing this dataset used to extract last
        available dates.

    Returns
    -------
    Monthly nominal exchange rates : pd.DataFrame
        Sell rate, monthly average and end of period.

    """
    if pipeline is not None:
        pipeline = Pipeline()
    if previous_data.empty:
        start_date = dt.datetime(1999, 12, 31)
    else:
        start_date = previous_data.index[-1]

    today = dt.datetime.now() - dt.timedelta(days=1)
    runs = (today - start_date).days // 360
    data = []
    base_url = urls["nxr_daily"]["dl"]["main"]
    if runs > 0:
        for i in range(1, runs + 1):
            from_ = (start_date + dt.timedelta(days=1)).strftime('%d/%m/%Y')
            to_ = (start_date + dt.timedelta(days=360)).strftime('%d/%m/%Y')
            dates = f"%22FechaDesde%22:%22{from_}%22,%22FechaHasta%22:%22{to_}"
            url = f"{base_url}{dates}%22,%22Grupo%22:%222%22}}" + "}"
            try:
                data.append(pd.read_excel(url))
                start_date = dt.datetime.strptime(to_, '%d/%m/%Y')
            except (TypeError, BadZipFile):
                pass
    from_ = (start_date + dt.timedelta(days=1)).strftime('%d/%m/%Y')
    to_ = (dt.datetime.now() - dt.timedelta(days=1)).strftime('%d/%m/%Y')
    dates = f"%22FechaDesde%22:%22{from_}%22,%22FechaHasta%22:%22{to_}"
    url = f"{base_url}{dates}%22,%22Grupo%22:%222%22}}" + "}"
    try:
        data.append(pd.read_excel(url))
    except (TypeError, BadZipFile):
        pass
    try:
        output = pd.concat(data, axis=0)
        output = output.pivot(index="Fecha", columns="Moneda",
                              values="Venta").rename_axis(None)
        output.index = pd.to_datetime(output.index, format="%d/%m/%Y",
                                      errors="coerce")
        output.sort_index(inplace=True)
        output.replace(",", ".", regex=True, inplace=True)
        output.columns = ["Tipo de cambio US$, Cable"]
        output = output.apply(pd.to_numeric, errors="coerce")

        metadata._set(output, area="Precios", currency="UYU/USD",
                      inf_adj="No", unit="-", seas_adj="NSA",
                      ts_type="-", cumperiods=1)
        output.columns = output.columns.set_levels(["-"], level=2)
        output.rename_axis(None, inplace=True)

    except ValueError as e:
        if str(e) == "No objects to concatenate":
            return previous_data

    return output


# The `_contains_nan` function needs to be monkey-patched to avoid an error
# when checking whether a Series is True
def _new_contains_nan(a, nan_policy='propagate'):
    policies = ['propagate', 'raise', 'omit']
    if nan_policy not in policies:
        raise ValueError("nan_policy must be one of {%s}" %
                         ', '.join("'%s'" % s for s in policies))
    try:
        with np.errstate(invalid='ignore'):
            # This [0] gets the value instead of the array, fixing the error
            contains_nan = np.isnan(np.sum(a))[0]
    except TypeError:
        try:
            contains_nan = np.nan in set(a.ravel())
        except TypeError:
            contains_nan = False
            nan_policy = 'omit'
            warnings.warn("The input array could not be properly checked for "
                          "nan values. nan values will be ignored.",
                          RuntimeWarning)

    if contains_nan and nan_policy == 'raise':
        raise ValueError("The input contains nan values")

    return contains_nan, nan_policy


stats._contains_nan = _new_contains_nan


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def cpi_measures(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """
    Get core CPI, Winsorized CPI, tradabe CPI, non-tradable CPI and residual
    CPI.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Monthly CPI measures : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()

    name = "cpi_measures"

    try:
        xls_10_14 = pd.ExcelFile(urls[name]["dl"]["2010-14"])
        xls_15 = pd.ExcelFile(urls[name]["dl"]["2015-"])
        prod_97 = (pd.read_excel(urls[name]["dl"]["1997"],
                                    skiprows=5)
                    .dropna(how="any")
                    .set_index(
                        "Rubros, Agrupaciones, Subrubros, Familias y Artículos")
                    .T)
    except URLError as err:
        if "SSL: CERTIFICATE_VERIFY_FAILED" in str(err):
            certificate = Path(get_project_root(), "utils", "files",
                               "ine_certs.pem")
            r = requests.get(urls[name]["dl"]["2010-14"],
                             verify=certificate)
            xls_10_14 = pd.ExcelFile(BytesIO(r.content))
            r = requests.get(urls[name]["dl"]["2015-"],
                             verify=certificate)
            xls_15 = pd.ExcelFile(BytesIO(r.content))
            r = requests.get(urls[name]["dl"]["1997"],
                             verify=certificate)
            prod_97 = (pd.read_excel(BytesIO(r.content),
                                     skiprows=5).dropna(how="any")
                       .set_index(
                "Rubros, Agrupaciones, Subrubros, Familias y Artículos")
                .T)
        else:
            raise err
    weights_97 = (pd.read_excel(urls[name]["dl"]["1997_weights"],
                                index_col=0)
                  .drop_duplicates(subset="Descripción", keep="first"))
    weights = pd.read_excel(xls_10_14, sheet_name=xls_10_14.sheet_names[0],
                            usecols="A:C", skiprows=13,
                            index_col=0).dropna(how="any")
    weights.columns = ["Item", "Weight"]
    weights_8 = weights.loc[weights.index.str.len() == 8]

    sheets = []
    for excel_file in [xls_10_14, xls_15]:
        for sheet in excel_file.sheet_names:
            raw = pd.read_excel(excel_file, sheet_name=sheet, usecols="D:IN",
                                skiprows=8).dropna(how="all")
            proc = raw.loc[:, raw.columns.str.
                           contains("Indice|Índice")].dropna(how="all")
            sheets.append(proc.T)
    complete_10 = pd.concat(sheets)
    complete_10 = complete_10.iloc[:, 1:]
    complete_10.columns = [weights["Item"], weights.index]
    complete_10.index = pd.date_range(start="2010-12-31",
                                      periods=len(complete_10), freq="M")
    diff_8 = complete_10.loc[:,
                             complete_10.columns.get_level_values(
                                 level=1).str.len()
                             == 8].pct_change()
    win = pd.DataFrame(winsorize(diff_8, limits=(0.05, 0.05), axis=1))
    win.index = diff_8.index
    win.columns = diff_8.columns.get_level_values(level=1)
    cpi_win = win.mul(weights_8.loc[:, "Weight"].T)
    cpi_win = cpi_win.sum(axis=1).add(1).cumprod().mul(100)

    weights_97["Weight"] = (weights_97["Rubro"]
                            .fillna(weights_97["Agrupación, subrubro, familia"])
                            .fillna(weights_97["Artículo"])
                            .drop(columns=["Rubro",
                                           "Agrupación, subrubro, familia",
                                           "Artículo"]))
    prod_97 = prod_97.loc[:, list(cpi_details["1997_base"].keys())]
    prod_97.index = pd.date_range(start="1997-03-31",
                                  periods=len(prod_97), freq="M")
    weights_97 = (weights_97[weights_97["Descripción"]
                             .isin(cpi_details["1997_weights"])]
                  .set_index("Descripción")
                  .drop(columns=["Rubro", "Agrupación, subrubro, "
                                          "familia", "Artículo"])).div(100)
    weights_97.index = prod_97.columns
    prod_10 = complete_10.loc[:, list(cpi_details["2010_base"].keys())]
    prod_10 = prod_10.loc[:, ~prod_10.columns.get_level_values(level=0)
                          .duplicated()]
    prod_10.columns = prod_10.columns.get_level_values(level=0)
    weights_10 = (weights.loc[weights["Item"]
                              .isin(list(cpi_details["2010_base"].keys()))]
                  .drop_duplicates(subset="Item",
                                   keep="first")).set_index("Item")
    items = []
    weights = []
    for item, weight, details in zip([prod_10, prod_97],
                                     [weights_10, weights_97],
                                     ["2010_base", "1997_base"]):
        for tradable in [True, False]:
            items.append(
                item.loc[:, [k for k, v in cpi_details[details].items()
                             if v["Tradable"] is tradable]])
            aux = weight.loc[[k for k, v in cpi_details[details].items()
                              if v["Tradable"] is tradable]]
            weights.append(aux.div(aux.sum()))
        for core in [True, False]:
            items.append(
                item.loc[:, [k for k, v in cpi_details[details].items()
                             if v["Core"] is core]])
            aux = weight.loc[[k for k, v in cpi_details[details].items()
                              if v["Core"] is core]]
            weights.append(aux.div(aux.sum()))

    intermediate = []
    for item, weight in zip(items, weights):
        intermediate.append(item.mul(weight.squeeze()).sum(1))

    output = []
    for x, y in zip(intermediate[:4], intermediate[4:]):
        aux = pd.concat([y.pct_change().loc[y.index < "2011-01-01"],
                         x.pct_change().loc[x.index > "2011-01-01"]])
        output.append(aux.fillna(0).add(1).cumprod().mul(100))

    pipeline.get("cpi")
    cpi_re = pipeline.dataset
    cpi_re = cpi_re.loc[cpi_re.index >= "1997-03-31"]
    output = pd.concat([cpi_re] + output + [cpi_win], axis=1)
    output.columns = ["Índice de precios al consumo: total",
                      "Índice de precios al consumo: transables",
                      "Índice de precios al consumo: no transables",
                      "Índice de precios al consumo: subyacente",
                      "Índice de precios al consumo: residual",
                      "Índice de precios al consumo: Winsorized 0.05"]
    output = output.apply(pd.to_numeric, errors="coerce")
    metadata._set(output, area="Precios y salarios", currency="-",
                  inf_adj="No", unit="2010-12=100", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    output = transform.rebase(output, start_date="2010-12-01",
                              end_date="2010-12-31")
    output.rename_axis(None, inplace=True)

    return output
